import math

from quicktions import Fraction


class ArithmeticProgressionError(BaseException):
    def __init__(self, *args):
        super().__init__(*args)


class DAndSError(ArithmeticProgressionError):
    def __init__(self, *args):
        msg = 'you cannot set both d an s!'
        super().__init__(msg, *args)


class ArithmeticProgression(object):
    def __init__(self, a1=None, an=None, n=None, d=None, s=None, correct_s=False):
        self._a1 = None
        self._an = None
        self._n = None
        self._d = None
        self._s = None
        self._current = None
        self._index = None
        self._correction_factor = None
        self._correct_s = None

        self.a1 = a1
        self.an = an
        self.n = n
        self.d = d
        self.s = s
        self.correct_s = correct_s

    def _to_fraction(self, value):
        if not isinstance(value, Fraction):
            value = Fraction(value)
        return value

    def _check_args(self, arg=None):
        if arg is None:
            err = 'Not enough attributes are set. Three are needed!'
            if len([v for v in self._parameters_dict.values() if v is not None]) < 3:
                raise AttributeError(err)
        else:
            if self._parameters_dict[arg] is None and len(
                    [v for v in self._parameters_dict.values() if v is not None]) > 2:
                err = 'attribute cannot be set. Three parameters are already set. ArithmeticProgression is already ' \
                      'created!'
                raise AttributeError(err)

    def _calculate_a1(self):
        if self._d is None:
            # self._a1 = (2. * self.s / self.n) - self.an
            self._a1 = Fraction(2 * self.s, self.n) - self.an
        else:
            self._a1 = self.an - ((self.n - 1) * self.d)

    def _calculate_an(self):
        if self._s is None:
            self._an = self.a1 + (self.n - 1) * self.d
        else:
            self._an = Fraction(2 * self.s, self.n) - self.a1
            # self._an = (2. * self.s / self.n) - self.a1

    def _calculate_n(self):
        if self._s is None:
            self._n = Fraction((self.an - self.a1), self.d) + 1
        else:
            self._n = 2 * Fraction(self.s, (self.a1 + self.an))
        self._n = Fraction(int(float(self._n)))

    def _calculate_d(self):
        if self.n == 1:
            self._d = 0
        elif self._a1 is None:
            self._calculate_a1()
            # self._d = Fraction(Fraction(self.an - self.a1), Fraction(self.n - 1))
            self._d = Fraction((self.an - self.a1), (self.n - 1))
            # self._d = (self.an - self.a1) / (self.n - 1)
        elif self._an is None:
            # self._d = Fraction(Fraction((self.s - (self.n * self.a1)) * 2), Fraction((self.n - 1) * self.n))
            self._d = Fraction(((self.s - (self.n * self.a1)) * 2), ((self.n - 1) * self.n))
            # self._d = ((self.s - (self.n * self.a1)) * 2) / ((self.n - 1) * self.n)
        elif self._n is None:
            self._calculate_n()
            # self._d = Fraction(Fraction(self.an - self.a1), Fraction(self.n - 1))
            self._d = Fraction((self.an - self.a1), (self.n - 1))
            # self._d = (self.an - self.a1) / (self.n - 1)
        else:
            self._d = Fraction((self.an - self.a1), (self.n - 1))
            # self._d = (self.an - self.a1) / (self.n - 1)

    def _calculate_s(self):
        if self._a1 is None:
            self._calculate_a1()
            self._s = (self.a1 + self.an) * Fraction(self.n, 2)
        elif self._an is None:
            self._s = self.n * self.a1 + ((self.n - 1) * Fraction(self.n, 2)) * self.d
        elif self._n is None:
            self._calculate_n()
            self._s = (self.a1 + self.an) * Fraction(self.n, 2)
        else:
            self._s = (self.a1 + self.an) * Fraction(self.n, 2)

    @property
    def _parameters_dict(self):
        return {'a1': self._a1, 'an': self._an, 'n': self._n, 'd': self._d, 's': self._s}

    @property
    def parameters_dict(self):
        return {'a1': self.a1, 'an': self.an, 'n': self.n, 'd': self.d, 's': self.s}

    @property
    def a1(self):
        if self._a1 is None:
            self._calculate_a1()
        return self._a1

    @a1.setter
    def a1(self, value):
        if value is not None:
            value = self._to_fraction(value)
            self._check_args('a1')
        self._a1 = value

    @property
    def an(self):
        if self._an is None:
            self._calculate_an()
        return self._an

    @an.setter
    def an(self, value):
        if value is not None:
            value = self._to_fraction(value)
            self._check_args('an')
        self._an = value

    @property
    def n(self):
        if self._n is None:
            self._calculate_n()
        return self._n

    @n.setter
    def n(self, value):
        if value is not None:
            if not isinstance(value, int):
                raise AttributeError('n {} must be int'.format(value))
            self._check_args('n')
        self._n = value

    @property
    def s(self):
        if self._s is None:
            self._calculate_s()
        return self._s

    @s.setter
    def s(self, value):
        if value is not None:
            value = self._to_fraction(value)
            self._check_args('s')
            if self._d is not None:
                raise DAndSError()
        self._s = value

    @property
    def d(self):
        if self._d is None:
            self._calculate_d()
        return self._d

    @d.setter
    def d(self, value):
        if value is not None:
            value = self._to_fraction(value)
            self._check_args('d')
            if self._s is not None:
                raise DAndSError()
        self._d = value

    @property
    def correct_s(self):
        return self._correct_s

    @correct_s.setter
    def correct_s(self, val):
        if not isinstance(val, bool):
            raise TypeError('correct_s.value must be of type bool not{}'.format(type(val)))
        self._correct_s = val

    @property
    def correction_factor(self):
        def _calculate_correction_factor():
            if self.correct_s:
                actual_s = self.n * Fraction((self.a1 + self.an), 2)
                factor = Fraction(self.s, actual_s)
                return factor
            else:
                return 1

        if self._correction_factor is None:
            self._correction_factor = _calculate_correction_factor()
        return self._correction_factor

    def __iter__(self):
        return self

    def __next__(self):
        if self._current is None:
            self._check_args()
        parameters = [self.a1, self.an, self.n, self.s, self.d]

        if len([v for v in parameters if v is not None]) < 5:
            err = 'Not enough parameter set to create an arithmetic progression. 3 parameters should be set first (s ' \
                  'and d cannot be set together) '
            raise Exception(err)

        if self._current is None:
            self._current = self.a1
            self._index = 0
        else:
            self._index += 1
            self._current += self.d

        if self._index < self.n:
            return self._current * self.correction_factor
        else:
            raise StopIteration()

    @property
    def index(self):
        return self._index

    # def rest(self):
    #     self._current = None

    def __deepcopy__(self, memodict={}):
        copy = self.__class__(correct_s=self.correct_s)
        copy._a1 = self._a1
        copy._an = self._an
        copy._n = self._n
        copy._d = self._d
        copy._s = self._s
        return copy



import os

from musurgia.pdf.text import PageText
from musurgia.pdf.positioned import Positioned
from prettytable import PrettyTable
from quicktions import Fraction

from musurgia.fractaltree.fractalmusicsquare import Module
from musurgia.timeline.timeline import TimeLine

path = os.path.abspath(__file__).split('.')[0]


class ModuleTimeLine(object):
    def __init__(self, start_time, module, instruments, text=None, number=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._start_time = None
        self.start_time = start_time
        self._instruments = None
        self.instruments = instruments
        self._module = None
        self.module = module
        self.text = text
        self.number = number

    @property
    def module(self):
        return self._module

    @module.setter
    def module(self, val):
        if self._module:
            raise Exception()
        if not isinstance(val, Module):
            raise TypeError('module.value must be of type Module not{}'.format(type(val)))
        self._module = val

    @property
    def start_time(self):
        return self._start_time

    @start_time.setter
    def start_time(self, val):
        if self._start_time:
            raise Exception()
        self._start_time = val

    @property
    def instruments(self):
        return self._instruments

    @instruments.setter
    def instruments(self, val):
        if self._instruments:
            raise Exception()
        self._instruments = val

    def get_end_time(self):
        return int(round(self.start_time + round(self.module.duration)))


class Vertical(Positioned):
    def __init__(self, parent, position, number=None, line_type='dashed', thickness=1, mode='start', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.position = position
        self.parent = parent
        self.mode = mode
        self.number = number
        self._line_type = None
        self.line_type = line_type
        self.thickness = thickness

    @property
    def line_type(self):
        return self._line_type

    @line_type.setter
    def line_type(self, val):
        permitted = ['dashed', 'normal']
        if val not in permitted:
            raise ValueError('line_type.value {} must be in {}'.format(val, permitted))
        self._line_type = val

    def draw(self, pdf):
        if self.mode == 'start':
            line = self.parent.ruler.line_segments[self.position]
            x = line.x1
        elif self.mode == 'end':
            line = self.parent.ruler.line_segments[self.position - 1]
            x = line.x2
        else:
            raise ValueError()

        y = line.y1
        p = line.page
        pdf.x = x
        pdf.y = y
        pdf.page = p

        page_text = PageText(self.number, relative_y=-7, font_size=10, font_weight='bold')
        page_text.draw(pdf)

        for i in range(self.thickness):
            grid = 0.1
            y1 = y + self.relative_y
            y2 = y

            if self.line_type == 'dashed':
                pdf.dashed_line(x1=x + i * grid, x2=x + i * grid, y1=y1, y2=y2 + self.parent.get_height() - 7,
                                space_length=3)
            else:
                pdf.line(x1=x + i * grid, x2=x + i * grid, y1=y1, y2=y2 + self.parent.get_height() - 7)


class ScoreTimeLine(TimeLine):
    def __init__(self, instruments, units_per_line=30, show_interval=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._instruments = None
        self.instruments = instruments
        self._module_time_lines = []
        self.units_per_line = units_per_line
        self._verticals = []
        self.show_interval = show_interval

    @property
    def instruments(self):
        return self._instruments

    @instruments.setter
    def instruments(self, val):
        self._instruments = val

    @property
    def module_time_lines(self):
        return self._module_time_lines

    def add_module_time_line(self, module_time_line):
        if not isinstance(module_time_line, ModuleTimeLine):
            raise TypeError()
        for instrument in module_time_line.instruments:
            if instrument not in self.instruments:
                raise ValueError('instrument {} not in {}'.format(instrument, self.instruments))
        # module_time_line.number = len(self._module_time_lines) + 1
        self._module_time_lines.append(module_time_line)

    def get_duration(self):
        return int(max([mtl.get_end_time() for mtl in self.module_time_lines]))

    def add_vertical(self, position, mode='start'):
        v = Vertical(parent=self, position=position, number=len(self._verticals) + 1, mode=mode)
        self._verticals.append(v)
        return v

    def apply_module_time_lines(self):
        self.length = self.get_duration()
        for instrument in self.instruments:
            voice = self.add_voice(name=instrument.abbreviation)
            voice.instrument = instrument
        for module_time_line in self.module_time_lines:
            voices = [v for v in self.voices if v.instrument in module_time_line.instruments]
            for voice in voices:
                module = module_time_line.module
                segment = voice.add_voice_segment(module_time_line.start_time, module_time_line.get_end_time())
                segment.lines[0].add_text_label(module.name, font_size=8, font_weight='bold', relative_x=1)
                try:
                    segment.lines[2].add_text_label('t=' + str(module.tempo), font_size=8)
                except IndexError:
                    segment.lines[0].add_text_label('t=' + str(module.tempo), font_size=4, y_offset=-4.5,
                                                            x_offset=4)
                try:
                    segment.lines[4].add_text_label('d=' + str(round(module.duration)) + '"', font_size=8)
                except IndexError:
                    segment.lines[0].add_text_label('d=' + str(round(module.duration)) + '"', font_size=4,
                                                            x_offset=8, y_offset=-4.5)
                if module_time_line.text:
                    segment.lines[0].add_text_label(module_time_line.text, font_size=4, y_offset=-4.5, x_offset=1)
            self.add_vertical(position=module_time_line.start_time)
            self.add_vertical(position=module_time_line.get_end_time(), mode='end')

    def apply_verticals(self, pdf):
        self._verticals.sort(key=lambda v: v.position)
        number = 1
        position = None
        for vertical in self._verticals:
            if vertical.position != position:
                if vertical == self._verticals[-1]:
                    vertical.thickness = 4
                    vertical.relative_y = -3
                    vertical.line_type = 'normal'

                vertical.number = number
                vertical.draw(pdf)
                position = vertical.position
                number += 1

    def draw(self, pdf):
        self.apply_module_time_lines()
        self.ruler.show_interval = self.show_interval
        printable = (pdf.w - pdf.r_margin - pdf.l_margin)
        self.unit = Fraction(Fraction(printable), Fraction(self.units_per_line))
        super().draw(pdf)
        self.apply_verticals(pdf)

    def draw_square(self, square_path, square):
        os.system('touch ' + square_path)
        file = open(square_path, 'w')
        x = PrettyTable(hrules=1)
        x.set_style(11)
        column_numbers = [str(number) for number in range(1, square.side_size + 1)]
        x.field_names = ["instrument", 'row', *column_numbers]
        for instrument in self.instruments:
            mtls = [mtl for mtl in self.module_time_lines if instrument in mtl.instruments]
            for row_number in range(1, square.side_size + 1):
                row_mtls = [mtl for mtl in mtls if mtl.module.row_number == row_number]
                dict_modules = {}
                for mtl in row_mtls:
                    module = mtl.module
                    try:
                        column_number = module.new_column_number
                    except AttributeError:
                        column_number = module.column_number
                    try:
                        dict_modules[column_number].append(mtl.number)
                    except KeyError:
                        dict_modules[column_number] = [mtl.number]

                row_infos = square.side_size * ['']
                for key in dict_modules.keys():
                    row_infos[key - 1] = str(dict_modules[key]).strip('[]')

                x.add_row([instrument.abbreviation, row_number, *row_infos])

        file.write(x.get_string())
        file.close()



import os
from itertools import cycle

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.chordfield.chordfield import ChordField
from musurgia.chordfield.valuegenerator import ValueGenerator

path = str(os.path.abspath(__file__).split('.')[0])
score = TreeScoreTimewise()
arth_durations = [2, 4]
new_child_durations = [1, 5]

arith_chord_field = ChordField()
for arth_duration in arth_durations:
    arith_chord_field.add_child(ChordField(quarter_duration=arth_duration,
                                           duration_generator=ValueGenerator(
                                               ArithmeticProgression(a1=1, an=0.25, correct_s=True))))

arith_chord_field.midi_generator = ValueGenerator(cycle([60]))
# copied_arith_chord_field = arith_chord_field.__deepcopy__()
# print([float(x) for x in (list(arith_chord_field.duration_generator.__deepcopy__()))])
# print(sum(list(arith_chord_field.duration_generator.__deepcopy__())))
new_chord_field = ChordField(duration_generator=arith_chord_field.duration_generator.__deepcopy__())
# new_chord_field = ChordField(duration_generator=arith_chord_field.duration_generator)
for i, new_child_duration in enumerate(new_child_durations):
    midi = 60 + i
    new_chord_field.add_child(ChordField(quarter_duration=new_child_duration,
                                         midi_generator=ValueGenerator(cycle([midi])),
                                         long_ending_mode='self_extend',
                                         short_ending_mode='self_shrink'))

arith_chord_field.simple_format.to_stream_voice().add_to_score(score=score, part_number=1)
# copied_arith_chord_field.simple_format.to_stream_voice().add_to_score(score=score, part_number=1)
simple_format = SimpleFormat()
for child in new_chord_field.children:
    simple_format.extend(child.simple_format)

simple_format.to_stream_voice().add_to_score(score=score, part_number=2)
xml_path = path + '.xml'
score.write(xml_path)



from pathlib import Path

from musurgia.pdf.line import HorizontalLineSegment, VerticalSegmentedLine
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.rowcolumn import DrawObjectRow, DrawObjectColumn
from musurgia.unittest import TestCase

path = Path(__file__)


class TestRowColumn(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf(orientation='l')
        self._make_row()
        self._make_column()

    def _make_row(self):
        r = DrawObjectRow()
        r.add_draw_object(HorizontalLineSegment(10))
        do2 = r.add_draw_object(HorizontalLineSegment(20))
        r.add_draw_object(VerticalSegmentedLine(lengths=[5, 6, 7, 8]))
        do2.start_mark_line.length = 6
        self.row = r

    def _make_column(self):
        c = DrawObjectColumn()
        c.add_draw_object(HorizontalLineSegment(10))
        do2 = c.add_draw_object(HorizontalLineSegment(20))
        c.add_draw_object(VerticalSegmentedLine(lengths=[5, 6, 7, 8]))
        do2.start_mark_line.length = 6
        self.column = c

    def test_draw_row(self):
        with self.file_path(path, 'draw_row', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            r = self.row
            r.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_row_of_segments(self):
        r = DrawObjectRow()
        r.add_draw_object(HorizontalLineSegment(30))
        r.add_draw_object(HorizontalLineSegment(10))
        r.add_draw_object(HorizontalLineSegment(20))
        with self.file_path(path, 'draw_row_of_segments', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            r.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_column_of_row_of_segments(self):
        r = DrawObjectRow()
        r.add_draw_object(HorizontalLineSegment(30))
        r.add_draw_object(HorizontalLineSegment(10))
        r.add_draw_object(HorizontalLineSegment(20))
        c = DrawObjectColumn()
        c.add_draw_object(HorizontalLineSegment(60))
        c.add_draw_object(r)

        with self.file_path(path, 'draw_column_of_row_of_segments', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            c.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_row_of_column_of_segments(self):
        c = DrawObjectColumn()
        c.add_draw_object(HorizontalLineSegment(30))
        c.add_draw_object(HorizontalLineSegment(10))
        c.add_draw_object(HorizontalLineSegment(20))
        r = DrawObjectRow()
        r.add_draw_object(c)

        with self.file_path(path, 'draw_row_of_column_of_segments', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            c.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_column(self):
        with self.file_path(path, 'draw_column', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            c = self.column
            c.draw(self.pdf)
            self.pdf.write(pdf_path)

    # def test_column_add_label_left(self):
    #     with self.file_path(path, 'column_add_label_left', 'pdf') as pdf_path:
    #         self.pdf.translate_page_margins()
    #         self.pdf.draw_ruler('h')
    #         self.pdf.draw_ruler('v')
    #         self.pdf.translate(20, 20)
    #         c = self.column
    #         c.add_label('test label', placement='left')
    #         c.draw(self.pdf)
    #         self.pdf.write(pdf_path)



import os

from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()
        fm = FractalMusic(tempo=60, quarter_duration=10, reading_direction='vertical')
        fm.add_layer()
        fm.add_layer()
        self.fm = fm.get_children()[1]
        self.deep_copied = self.fm.__deepcopy__()

    def test(self, exp=None, act=None):
        if not exp:
            exp = self.fm

        if not act:
            act = self.deep_copied

        self.assertEqual(exp.value, act.value)
        self.assertEqual(exp.proportions, act.proportions)
        self.assertEqual(exp.value, act.value)
        self.assertEqual(exp.proportions, act.proportions)
        self.assertEqual(exp.tree_permutation_order, act.tree_permutation_order)
        self.assertEqual(exp.fractal_order, act.fractal_order)
        self.assertEqual(exp.reading_direction, act.reading_direction)
        self.assertEqual(exp._name, act._name)

        self.assertEqual(exp.tree_directions, act.tree_directions)
        self.assertEqual(exp.tempo, act.tempo)
        self.assertEqual(exp.midi_value, act.midi_value)

    def test_1(self):
        self.assertIsNone(self.deep_copied.up)

    def test_2(self):
        self.assertNotEqual(self.deep_copied.name, self.fm.name)

    def test_3(self):
        for leaf in self.fm.traverse_leaves():
            leaf.chord.add_words(leaf.fractal_order)

        copied = self.fm.__deepcopy__()
        copied.get_simple_format().to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_3.xml'
        self.score.write(xml_path)
        expected_path = path + '_test_3_expected.xml'
        expected_score = TreeScoreTimewise()
        self.fm.get_simple_format().to_stream_voice().add_to_score(expected_score)
        expected_score.write(expected_path)
        self.assertCompareFiles(xml_path, expected_path)



from musurgia.pdf.newdrawobject import DrawObject
from musurgia.pdf.labeled import Labeled
from musurgia.pdf.named import Named
from musurgia.pdf.positioned import RelativeXNotSettableError, RelativeYNotSettableError


class MarkLine(DrawObject, Labeled):
    def __init__(self, parent, placement='start', height=3, thickness=1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.left_margin, self.top_margin, self.right_margin, self.bottom_margin = 0, 0, 0, 0
        self._parent = None
        self._height = None
        self._placement = None
        self.parent = parent
        self.placement = placement
        self.height = height
        self.position = None
        self.thickness = thickness

    @property
    def relative_x(self):
        if self.placement == 'start':
            return self.parent.relative_x
        else:
            return self.parent.relative_x + self.parent.actual_length

    @relative_x.setter
    def relative_x(self, val):
        if val is not None:
            raise RelativeXNotSettableError()

    @property
    def relative_y(self):
        return self.parent.relative_y

    @relative_y.setter
    def relative_y(self, val):
        if val is not None:
            raise RelativeYNotSettableError()

    @property
    def placement(self):
        return self._placement

    @placement.setter
    def placement(self, val):
        permitted = ['start', 'end']
        if val not in permitted:
            raise ValueError('placement.value {} must be in {}'.format(val, permitted))
        self._placement = val

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, val):
        self._height = val

    def get_relative_x2(self):
        return self.relative_x

    def get_relative_y2(self):
        return self.relative_y + self.height

    def draw(self, pdf):
        x1 = self.relative_x + pdf.x
        x2 = x1
        y1 = self.relative_y + pdf.y - self.height / 2
        y2 = y1 + self.height
        if self.show:
            for i in range(self.thickness):
                grid = 0.1
                pdf.line(x1=x1 + i * grid, y1=y1, x2=x2 + i * grid, y2=y2)
        self.position = (x1, y2)
        for text_label in self.text_labels:
            text_label.draw(pdf)
        # pdf.x = x2
        # pdf.y = y2


class LineSegment(DrawObject, Labeled, Named):
    def __init__(self, length, factor=1, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._length = None
        self._factor = None
        self._parent = None
        self.length = length
        self.factor = factor
        self._start_mark_line = MarkLine(parent=self, placement='start')
        self._end_mark_line = MarkLine(parent=self, placement='end', show=False)
        self._x1 = None
        self._x2 = None
        self._y1 = None
        self._y2 = None
        self._page = None

    @property
    def x1(self):
        return self._x1

    @property
    def x2(self):
        return self._x2

    @property
    def y1(self):
        return self._y1

    @property
    def y2(self):
        return self._y2

    @property
    def page(self):
        return self._page

    @property
    def factor(self):
        if self._factor is None:
            self._factor = 1
        return self._factor

    @factor.setter
    def factor(self, val):
        self._factor = val

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, val):
        self._length = val

    @property
    def actual_length(self):
        return self.length * self.factor

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, val):
        if not isinstance(val, SegmentedLine):
            raise TypeError(f"parent.value must be of type SegmentedLine not{type(val)}")
        self._parent = val

    @property
    def start_mark_line(self):
        return self._start_mark_line

    @property
    def end_mark_line(self):
        return self._end_mark_line

    def get_relative_x2(self):
        return self.relative_x + self.actual_length

    def get_relative_y2(self):
        return max([self.start_mark_line.get_relative_y2(), self.end_mark_line.get_relative_y2()])

    def draw(self, pdf):
        x1 = pdf.x + self.relative_x
        x2 = pdf.x + self.relative_x + self.actual_length
        y1 = pdf.y + self.relative_y
        y2 = y1

        self._x1, self._x2, self._y1, self._y2 = x1, x2, y1, y2
        self._page = pdf.page
        if self.name:
            self.name.draw(pdf)

        if self.show:
            self.start_mark_line.draw(pdf)
            self.end_mark_line.draw(pdf)
            pdf.line(x1=x1, y1=y1, x2=x2, y2=y2)
            for text_label in self._text_labels:
                text_label.draw(pdf)

        pdf.x = x2
        pdf.y += self.bottom_margin


class SegmentedLine(DrawObject, Labeled, Named):
    def __init__(self, lengths, bottom_margin=10, factor=1, *args, **kwargs):
        super().__init__(bottom_margin=bottom_margin, *args, **kwargs)
        self._line_segments = []
        self._factor = None
        self.factor = factor
        self._lengths = None
        self.lengths = lengths

    @DrawObject.relative_y.setter
    def relative_y(self, val):
        self._relative_y = val
        try:
            for line in self.line_segments:
                line.relative_y = val
        except AttributeError:
            pass

    @DrawObject.relative_x.setter
    def relative_x(self, val):
        self._relative_x = val
        try:
            self.line_segments[-1].relative_x = val
        except AttributeError:
            pass

    @property
    def lengths(self):
        return self._lengths

    @lengths.setter
    def lengths(self, val):
        if not isinstance(val, list):
            raise TypeError('lengths.value must be of type list not{}'.format(type(val)))
        self._lengths = val
        self._generate_line_segments()

    @property
    def line_segments(self):
        return self._line_segments

    def _generate_line_segments(self):
        self._line_segments = []
        for length in self.lengths:
            line_segment = LineSegment(length=length, relative_y=self.relative_y, factor=self.factor)

            if not self._line_segments:
                line_segment.relative_x = self.relative_x
            else:
                line_segment.relative_x = 0
            line_segment.parent = self
            self._line_segments.append(line_segment)
        self._line_segments[-1].end_mark_line.show = True

    def get_relative_x2(self):
        raise Exception('SegementedLine does not have a x2 value')

    def get_relative_y2(self):
        return self.line_segments[0].get_relative_y2()

    def draw_with_break(self, pdf):
        for text_label in self._text_labels:
            text_label.draw(pdf)

        for line_segment in self.line_segments:
            if line_segment == self.line_segments[0]:
                line_segment.name = self.name

            line_segment.draw_with_break(pdf)
            new_x = pdf.x

            if line_segment._line_break:
                # pdf.y += self.bottom_margin
                if self.name:
                    line_segment.name = self.name
                    pdf.x = new_x - line_segment.actual_length
                    line_segment.name.draw(pdf)
                pdf.x = new_x

    def draw(self, pdf):
        for text_label in self._text_labels:
            text_label.draw(pdf)
        pdf.x = pdf.l_margin
        for line_segment in self.line_segments:
            if line_segment == self.line_segments[0]:
                line_segment.name = self.name

            line_segment.draw_with_break(pdf)
            new_x = pdf.x

            if line_segment._line_break:
                # pdf.y += self.bottom_margin
                if self.name:
                    line_segment.name = self.name
                    pdf.x = new_x - line_segment.actual_length
                    line_segment.name.draw(pdf)
                pdf.x = new_x



from unittest import TestCase

from musurgia.fractaltree.fractaltree import FractalTree


class Test(TestCase):
    def setUp(self) -> None:
        self.ft = FractalTree(tree_permutation_order=[3, 1, 2], proportions=[1, 2, 3], value=10)

    def test_1(self):
        self.ft.add_layer()
        result = [(2, 1), (2, 2), (2, 3)]
        self.assertEqual(result, [child.multi for child in self.ft.get_children()])

    def test_2(self):
        self.ft.multi = (1, 7)
        self.ft.add_layer()
        result = [leaf.value for leaf in self.ft.traverse_leaves()]
        ft = FractalTree(value=10, multi=(3, 1))
        ft.add_layer()
        expected = [leaf.value for leaf in ft.traverse_leaves()]
        self.assertEqual(expected, result)

    def test_3(self):
        self.ft.multi = (4, 1)
        self.ft.add_layer()
        result = [leaf.value for leaf in self.ft.traverse_leaves()]
        ft = FractalTree(value=10, multi=(1, 1))
        ft.add_layer()
        expected = [leaf.value for leaf in ft.traverse_leaves()]
        self.assertEqual(expected, result)

    def test_4(self):
        self.ft.multi = (3, 4)
        self.ft.add_layer()
        result = [leaf.value for leaf in self.ft.traverse_leaves()]
        ft = FractalTree(value=10, multi=(1, 1))
        ft.add_layer()
        expected = [leaf.value for leaf in ft.traverse_leaves()]
        self.assertEqual(expected, result)



import os

from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic, SetTempoFirstException

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):

    def setUp(self) -> None:
        self.fm = FractalMusic(tempo=60, quarter_duration=12, tree_permutation_order=(3, 1, 2), proportions=[1, 2, 3])
        self.fm.midi_generator.set_directions(1, 1, -1)
        self.fm.midi_generator.midi_range = [55, 72]

    def test_1(self):
        fm = self.fm
        fm.add_layer()

        fm.add_layer(lambda n: True if n.fractal_order > 1 else False)
        fm.add_layer(lambda n: True if n.fractal_order > 1 else False)
        fm.add_layer(lambda n: True if n.fractal_order > 1 else False)
        # for node in fm.traverse():
        #     node.chord.add_words(node.midi_generator.midi_range)
        fm.add_layer(lambda n: True if n.fractal_order > 1 else False)

        # for node in fm.traverse():
        #     node.chord.add_words(node.children_generated_midis)
        #     node.chord.add_words(node.midi_generator.directions, relative_y=30)
        #     node.chord.add_words(node.children_generated_midis)
        #     node.chord.add_words(node.permutation_order, relative_y=60)

        score = TreeScoreTimewise()
        score = fm.get_score(score=score, show_fractal_orders=False)

        text_path = path + '_test_1.txt'
        fm.write_infos(text_path)
        self.assertCompareFiles(actual_file_path=text_path)

        xml_path = path + '_test_1.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)

    def test_2(self):
        fm = self.fm
        fm.add_layer()
        score_1 = fm.get_children()[0].get_score()
        xml_path = path + '_test_2.xml'
        score_1.write(path=xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)

    def test_3(self):
        fm = FractalMusic(duration=30)
        fm.add_layer()
        with self.assertRaises(SetTempoFirstException):
            fm.get_score()

    def test_4(self):
        fm = FractalMusic(duration=30)
        fm.add_layer()
        fm.get_children()[1].tempo = 72
        fm.set_none_tempi(80)
        fm.quantize_leaves(0.5)
        fm.add_layer()
        score = fm.get_root_score(layer_number=1, show_fractal_orders=True)
        xml_path = path + '_test_4.xml'
        score.write(path=xml_path)
        with self.assertRaises(SetTempoFirstException):
            fm.get_score()



import os
from unittest import TestCase

from musicscore.musictree.midi import B, A, C, F
from musicscore.musictree.treechord import TreeChord

from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):

    def test_1(self):
        fm = FractalMusic(tempo=60, quarter_duration=10, proportions=(1, 2, 3), tree_permutation_order=(3, 1, 2))
        fm.add_layer()
        midis = [B(4), A(4, 'b'), C(4, '#'), A(3), F(4)]
        fm.get_children()[0].add_layer()
        for leaf, midi in zip(fm.traverse_leaves(), midis):
            leaf.set_chord(TreeChord(midi))
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_.xml'
        score.write(path=xml_path)



from unittest import TestCase

from musurgia.fractaltree.fractalmusicsquare import Square


class Test(TestCase):
    def setUp(self) -> None:
        self.square = Square(duration=100, proportions=(1, 2), tree_permutation_order=(2, 1))
        self.copied_square = self.square.__deepcopy__()

    def test_1(self):
        self.assertNotEqual(self.square, self.copied_square)

    def test_2(self):
        self.assertNotEqual(self.square.rows, self.copied_square.rows)

    def test_3(self):
        self.assertNotEqual(self.square.get_all_modules(), self.copied_square.get_all_modules())

    def test_4(self):
        first_module = self.square.get_module(1, 1)
        old_duration = first_module.duration
        first_module.duration = 20
        self.assertEqual(old_duration, self.copied_square.get_module(1, 1).duration)

    def test_5(self):
        self.square.get_module(1, 1).duration = 20
        copied_square = self.square.__deepcopy__()
        self.assertEqual(20, copied_square.get_module(1, 1).duration)

    def test_6(self):
        for row in self.square.rows:
            row.set_name('L {} L'.format(row.number))
        copied_square = self.square.__deepcopy__()
        expected = [row.name for row in self.square.rows]
        result = [row.name for row in copied_square.rows]
        self.assertEqual(expected, result)



from pathlib import Path

from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from quicktions import Fraction

from musurgia.chordfield.chordfield import HalfWave, Wave
from musurgia.unittest import TestCase

path = Path(__file__)


class TestWave(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_half_up(self):
        w = HalfWave(quarter_duration=10,
                     proportions=[7, 5, 3, 1],
                     duration_units=[Fraction(1, 7), Fraction(1, 6), Fraction(1, 5), Fraction(1, 4)],
                     min_midi=60,
                     max_midi=76,
                     show_last_note=True)
        w.simple_format.to_stream_voice().add_to_score(self.score)
        with self.file_path(path, 'half_up', 'xml') as xml_path:
            self.score.write(xml_path)

    def test_half_down(self):
        w = HalfWave(quarter_duration=10,
                     proportions=[7, 5, 3, 1],
                     duration_units=[Fraction(1, 7), Fraction(1, 6), Fraction(1, 5), Fraction(1, 4)],
                     min_midi=60,
                     max_midi=76,
                     direction='down',
                     show_last_note=True)
        w.simple_format.to_stream_voice().add_to_score(self.score)
        with self.file_path(path, 'half_down', 'xml') as xml_path:
            self.score.write(xml_path)

    def test_simple(self):
        w = Wave(quarter_duration=13, field_proportions=[1, 1], show_last_note=True)
        w.rising_field.proportions = [5, 4, 3, 2, 1]
        w.falling_field.proportions = [5, 4, 3, 2, 1]
        w.simple_format.to_stream_voice().add_to_score(self.score)
        with self.file_path(path, 'simple', 'xml') as xml_path:
            self.score.write(xml_path)

    def test_simple_down(self):
        w = Wave(quarter_duration=13, field_proportions=[1, 1], show_last_note=True, direction='down')
        w.rising_field.proportions = [5, 3, 2, 1]
        w.falling_field.proportions = [5, 3, 2, 1]
        w.simple_format.to_stream_voice().add_to_score(self.score)
        with self.file_path(path, 'simple_down', 'xml') as xml_path:
            self.score.write(xml_path)

    def test_units(self):
        w = Wave(quarter_duration=13, field_proportions=[1, 1], show_last_note=True)
        w.rising_field.proportions = [5, 4, 3, 2, 1]
        w.rising_field.duration_units = [Fraction(1, 7), Fraction(1, 6), Fraction(1, 5), Fraction(1, 4)]
        w.falling_field.proportions = [5, 4, 3, 2, 1]
        w.falling_field.duration_units = w.rising_field.duration_units
        w.simple_format.to_stream_voice().add_to_score(self.score)
        with self.file_path(path, 'units', 'xml') as xml_path:
            self.score.write(xml_path)



from abc import abstractmethod

from quicktions import Fraction

from musurgia.pdf.labeled import Labeled
from musurgia.pdf.masterslave import Master, Slave
from musurgia.pdf.newdrawobject import DrawObject
from musurgia.pdf.rowcolumn import DrawObjectRow, DrawObjectColumn, DrawObjectContainer
from musurgia.pdf.text import TextLabel


class StraightLine(Slave, DrawObject, Labeled):
    def __init__(self, mode, length, show=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mode = None
        self._length = None

        self.mode = mode
        self.length = length
        self.show = show

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, val):
        permitted = ['h', 'horizontal', 'v', 'vertical']
        if val not in permitted:
            raise ValueError(f'mode.value {val} must be in {permitted}')
        self._mode = val

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, val):
        if not isinstance(val, float) and not isinstance(val, int) and not isinstance(val, Fraction):
            raise TypeError(f"length.value must be of type float, int or Fraction  not{type(val)}")
        self._length = val

    @property
    def is_vertical(self):
        if self.mode in ['v', 'vertical']:
            return True
        else:
            return False

    @property
    def is_horizontal(self):
        if self.mode in ['h', 'horizontal']:
            return True
        else:
            return False

    @staticmethod
    def get_opposite_mode(mode):
        if mode == 'h':
            return 'v'
        elif mode == 'v':
            return 'h'
        elif mode == 'horizontal':
            return 'vertical'
        elif mode == 'vertical':
            return 'horizontal'
        else:
            raise AttributeError()

    def get_relative_x2(self):
        if self.mode in ['h', 'horizontal']:
            return self.relative_x + self.length
        else:
            return self.relative_x

    def get_relative_y2(self):
        if self.mode in ['v', 'vertical']:
            return self.relative_y + self.length
        else:
            return self.relative_y

    def draw(self, pdf):
        if self.show:
            with pdf.prepare_draw_object(self):
                self.draw_above_text_labels(pdf)
                self.draw_left_text_labels(pdf)
                x2 = self.get_relative_x2() - self.relative_x
                y2 = self.get_relative_y2() - self.relative_y
                pdf.line(0, 0, x2, y2)
                self.draw_below_text_labels(pdf)


class MarkLine(StraightLine):
    def __init__(self, placement, length=3, *args, **kwargs):
        super().__init__(length=length, *args, **kwargs)
        self._placement = None
        self.placement = placement

    @property
    def placement(self):
        return self._placement

    @placement.setter
    def placement(self, val):
        permitted = ['start', 'end']
        if val not in permitted:
            raise ValueError(f'placement.value {val} must be in {permitted}')
        self._placement = val

    def get_middle_y(self):
        return self.length / 2


class LineSegment(Master, DrawObject):
    def __init__(self, mode, length, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._straight_line = StraightLine(name='straight_line', mode=mode, length=length, master=self)
        marker_mode = StraightLine.get_opposite_mode(self.mode)
        self._start_mark_line = MarkLine(name='start_mark_line', mode=marker_mode, master=self, placement='start')
        self._end_mark_line = MarkLine(name='end_mark_line', mode=marker_mode, master=self, placement='end', show=False)

    @property
    def straight_line(self):
        return self._straight_line

    @property
    def start_mark_line(self):
        return self._start_mark_line

    @property
    def end_mark_line(self):
        return self._end_mark_line

    @property
    def mode(self):
        return self.straight_line.mode

    @property
    def length(self):
        return self.straight_line.length

    def _get_straight_line_margin(self, margin):
        if margin in ['l', 'left']:
            return 0
        elif margin in ['r', 'right']:
            return 0
        elif margin in ['t', 'top']:
            return 0
        elif margin in ['b', 'bottom']:
            return 0
        else:
            raise AttributeError(margin)

    def _get_mark_line_margin(self, margin):
        if margin in ['l', 'left']:
            return 0
        elif margin in ['r', 'right']:
            return 0
        elif margin in ['t', 'top']:
            return 0
        elif margin in ['b', 'bottom']:
            return 0
        else:
            raise AttributeError(margin)

    def _get_straight_line_position(self, position):
        if position not in ['x', 'y']:
            raise AttributeError(position)

        if self.mode in ['h', 'horizontal']:
            if position == 'x':
                return 0
            elif position == 'y':
                return 0
                # return max([ml.get_middle_y() for ml in [self.start_mark_line, self.end_mark_line]])
        else:
            if position == 'x':
                return 0
                # return max([ml.get_middle_y() for ml in [self.start_mark_line, self.end_mark_line]])
            elif position == 'y':
                return 0

    def _get_mark_line_position(self, position, mark_line):
        if position not in ['x', 'y']:
            raise AttributeError(position)

        if mark_line.mode in ['h', 'horizontal']:
            if position == 'x':
                # return 0
                return -mark_line.length / 2
            else:
                if mark_line.placement == 'start':
                    return 0
                else:
                    return self.length
        else:
            if position == 'y':
                # return 0
                return -mark_line.length / 2
            else:
                if mark_line.placement == 'start':
                    return 0
                else:
                    return self.length

    def get_slave_margin(self, slave, margin):
        if slave.name == 'straight_line':
            return self._get_straight_line_margin(margin)
        elif slave.name == 'start_mark_line':
            return self._get_mark_line_margin(margin)
        elif slave.name == 'end_mark_line':
            return self._get_mark_line_margin(margin)
        else:
            raise AttributeError(slave)

    def get_slave_position(self, slave, position):
        if slave.name == 'straight_line':
            return self._get_straight_line_position(position)
        elif slave.name == 'start_mark_line':
            return self._get_mark_line_position(position, slave)
        elif slave.name == 'end_mark_line':
            return self._get_mark_line_position(position, slave)
        else:
            raise AttributeError(slave)


class HorizontalLineSegment(LineSegment):
    def __init__(self, length, *args, **kwargs):
        super().__init__(mode='horizontal', length=length, *args, **kwargs)

    def get_relative_x2(self):
        return self.relative_x + self.length

    def get_relative_y2(self):
        return self.relative_y + max([ml.get_height() for ml in [self.start_mark_line, self.end_mark_line]])

    def draw(self, pdf):
        with pdf.prepare_draw_object(self):
            self.start_mark_line.draw(pdf)
            self.straight_line.draw(pdf)
            self.end_mark_line.draw(pdf)


class VerticalLineSegment(LineSegment):
    def __init__(self, length, *args, **kwargs):
        super().__init__(mode='vertical', length=length, *args, **kwargs)

    def get_relative_x2(self):
        return self.relative_x + max([ml.get_width() for ml in [self.start_mark_line, self.end_mark_line]])

    def get_relative_y2(self):
        return self.relative_y + self.length

    def draw(self, pdf):
        with pdf.prepare_draw_object(self):
            self.start_mark_line.draw(pdf)
            self.straight_line.draw(pdf)
            self.end_mark_line.draw(pdf)


class AbstractSegmentedLine(DrawObjectContainer):
    def __init__(self, lengths, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._make_segments(lengths)

    @property
    def lengths(self):
        return [segment.length for segment in self.segments]

    @property
    def segments(self):
        return self.draw_objects

    @abstractmethod
    def _make_segments(self, lengths):
        pass


class HorizontalSegmentedLine(AbstractSegmentedLine, DrawObjectRow):

    def _make_segments(self, lengths):
        if not lengths:
            raise AttributeError('lengths must be set.')
        for length in lengths:
            self.add_draw_object(HorizontalLineSegment(length))
        self.segments[-1].end_mark_line.show = True


class VerticalSegmentedLine(AbstractSegmentedLine, DrawObjectColumn):

    def _make_segments(self, lengths):
        if not lengths:
            raise AttributeError('lengths must be set.')
        for length in lengths:
            self.add_draw_object(VerticalLineSegment(length))
        self.segments[-1].end_mark_line.show = True


class AbstractRuler(AbstractSegmentedLine):
    def __init__(self, length, unit=10.0, first_label=0, label_show_interval=1, show_first_label=True, *args, **kwargs):
        number_of_units = length / unit
        partial_segment_length = number_of_units - int(number_of_units)
        lengths = int(number_of_units) * [unit]
        if partial_segment_length:
            lengths += [partial_segment_length * unit]
        super().__init__(lengths=lengths, *args, **kwargs)
        if partial_segment_length:
            self.draw_objects[-1].end_mark_line.show = False
        self.unit = unit
        self.first_label = first_label
        self.label_show_interval = label_show_interval
        self.show_first_label = show_first_label
        self._set_labels()

    @property
    def length(self):
        return sum([s.length for s in self.segments])

    def _set_labels(self):
        def _add_label(mark_line, txt):
            tl = TextLabel(txt)
            if isinstance(self, VerticalSegmentedLine):
                tl.placement = 'left'
                tl.right_margin = 1
                tl.top_margin = 0
            else:
                tl.bottom_margin = 1
            mark_line.add_text_label(tl)

        for index, segment in enumerate(self.segments):
            if not self.show_first_label and index == 0:
                pass
            else:
                if index % self.label_show_interval == 0:
                    mark_line = segment.start_mark_line
                    _add_label(mark_line, index + self.first_label)

        last_segment_end_mark_line = self.segments[-1].end_mark_line
        if last_segment_end_mark_line.show and (len(self.segments)) % self.label_show_interval == 0:
            _add_label(last_segment_end_mark_line, len(self.segments) + self.first_label)


class HorizontalRuler(AbstractRuler, HorizontalSegmentedLine):
    pass


class VerticalRuler(AbstractRuler, VerticalSegmentedLine):
    pass



import os

from musurgia.fractaltree.fractalmusic import FractalMusic
from musurgia.unittest import TestCase

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.fm = FractalMusic(proportions=[1, 2, 3, 4], tree_permutation_order=[3, 1, 4, 2], duration=10.4)

    def test_1(self):
        self.fm.tempo = self.fm.find_best_tempo_in_range()
        self.assertEqual(10.4, self.fm.duration)

    def test_2(self):
        self.fm.tempo = self.fm.find_best_tempo_in_range()
        self.fm.change_quarter_duration(round(self.fm.quarter_duration))
        self.assertEqual(13, self.fm.quarter_duration)

    def test_3(self):
        self.fm.add_layer()
        for child in self.fm.get_children():
            best_tempo = child.find_best_tempo_in_range()
            child.tempo = best_tempo
        # print([leaf.quarter_duration for leaf in self.fm.get_children()])
        self.fm.round_leaves()
        result = [4.0, 1.0, 7.0, 2.0]
        self.assertEqual(result, [child.quarter_duration for child in self.fm.get_children()])

    def test_4(self):
        self.fm.add_layer()
        for child in self.fm.get_children():
            best_tempo = child.find_best_tempo_in_range()
            child.tempo = best_tempo
        self.fm.round_leaves()
        xml_path = path + '_test_4.xml'
        score = self.fm.get_children_score()
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_5(self):
        best_tempi = self.fm.find_best_tempi_in_list(tempi=[50, 60, 70, 80, 90, 100])
        self.assertEqual([70, 80], best_tempi)




from musurgia.random import Random


class ReadAList(object):
    ##mode in forwards, backwards, zickzack, random
    def __init__(self, pool=None, mode='random', seed=None):
        self._pool = None
        self._mode = None
        self._random = None
        self._index = None
        self._direction = 1
        self._next_index = None

        self.pool = pool
        self.mode = mode
        self.seed = seed

    @property
    def pool(self):
        return self._pool

    @pool.setter
    def pool(self, values):
        if values is not None:
            try:
                self._pool = list(values)
            except:
                self._pool = [values]
        self.random.pool = self.pool

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value):
        if value not in ['forwards', 'backwards', 'zickzack', 'random']:
            err = 'mode can only be forwards, backwards, zickzack or random'
            raise ValueError(err)
        self._mode = value

    @property
    def random(self):
        if self._random is None:
            self._random = Random()
        return self._random

    @property
    def seed(self):
        return self.random.seed

    @seed.setter
    def seed(self, value):
        self.random.seed = value

    @property
    def next_index(self):
        err = 'next_index can only be set'
        raise AttributeError(err)

    @next_index.setter
    def next_index(self, value):
        self._next_index = value

    def _set_next_index(self):

        if self.mode == 'forwards':
            self._direction = 1

        elif self.mode == 'backwards':
            self._direction = -1

        elif self.mode == 'zickzack':
            pass

        self._index += self._direction

    def _check_index(self):
        if self.mode == 'forwards':
            if self._index >= len(self.pool):
                self._index = 0

        elif self.mode == 'backwards':
            if self._index >= len(self.pool):
                self._index = len(self.pool) - 1
            elif self._index < 0:
                self._index = len(self.pool) - 1

        elif self.mode == 'zickzack':
            if self._index == len(self.pool) - 1:
                self._direction = -1

            elif self._index > len(self.pool) - 1:
                self._index = len(self.pool) - 1
                self._direction = -1

            elif self._index == 0:
                self._direction = 1

            elif self._index < 0:
                self._index = 1
                self._direction = 1

    def next(self):
        if self.pool is None:
            err = 'pool can not be None'
            raise AttributeError(err)

        if self.mode != 'random':
            # print 'read_a_list.next(): self.mode=',self.mode
            # print 'read_a_list.next(): self._next_index=',self._next_index

            if self._next_index is None and self._index is None:
                if self.mode == 'backwards':
                    self._next_index = len(self.pool) - 1
                else:
                    self._next_index = 0

            if self._next_index is None:
                self._set_next_index()
            else:
                self._index = self._next_index
                self._next_index = None

            self._check_index()
            # print 'read_a_list.next(): self._index after check=',self._index
            # print 'read_a_list.next(): self.pool', self.pool
            return self.pool[self._index]

        else:
            return self.random.next()



import os

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):

    def test_1(self):
        fm = FractalMusic(tempo=60, quarter_duration=10)
        fm.midi_generator.midi_range = [60, 72]
        fm.add_layer()
        fm.add_info('fractal_order')
        score = fm.get_score(layer_number=fm.number_of_layers)
        xml_path = path + '_test_1.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(xml_path)

    def test_2(self):
        fm = FractalMusic(tempo=60, quarter_duration=10)
        fm.midi_generator.midi_range = [60, 72]
        fm.add_layer()
        fm.add_layer()
        fm.add_info('fractal_order')
        score = fm.get_score()
        xml_path = path + '_test_2.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(xml_path)

    def test_3(self):
        fm = FractalMusic(tempo=60, quarter_duration=10)
        fm.midi_generator.midi_range = [60, 72]
        fm.add_layer()
        fm.add_layer()
        fm.add_info('midi_value')
        score = fm.get_score()
        xml_path = path + '_test_3.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(xml_path)

    def test_4(self):
        fm = FractalMusic(tempo=60, quarter_duration=10)
        fm.midi_generator.midi_range = [60, 72]
        fm.add_layer()
        fm.add_layer()
        fm.add_info((lambda node: int(node.midi_value) if int(node.midi_value) == node.midi_value else node.midi_values,
                     'above'))
        score = fm.get_score()
        xml_path = path + '_test_4.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(xml_path)



from itertools import cycle


def permute(input_list, permutation_order):
    if input_list is None:
        raise ValueError('input_list cannot be None')
    if permutation_order is None:
        raise ValueError('permutation_order cannot be None')
    if len(input_list) == len(permutation_order):
        return [input_list[m - 1] for m in permutation_order]
    else:
        raise ValueError('input_list = {} and permutation_order {} must have the same length'.format(str(input_list),
                                                                                                     str(
                                                                                                         permutation_order)))


def self_permute(permutation_order):
    output = [permutation_order]

    for i in range(1, len(permutation_order)):
        output.append(permute(output[i - 1], permutation_order))

    return output


def get_self_multiplied_permutation(permutation_order):
    self_permuted_order = self_permute(permutation_order)
    multiplied = [permute(self_permuted_order, current_order) for current_order in self_permuted_order]
    return multiplied


def get_reordered_self_multiplied_permutation(permutation_order):
    def _reorder(matrix):
        index_of_first_row = None
        for index_of_first_row in range(len(matrix)):
            if matrix[index_of_first_row][0] == permutation_order:
                break
        output = []
        for i in range(index_of_first_row, index_of_first_row + len(matrix)):
            output.append(matrix[i % len(matrix)])
        return output

    multiplied = get_self_multiplied_permutation(permutation_order)
    return _reorder(multiplied)


def get_vertical_self_multiplied_permutation(permutation_order):
    def _transpose(matrix):
        output = []
        for i in range(len(matrix)):
            temp = []
            for j in range(len(matrix)):
                temp2 = []
                for k in range(len(matrix)):
                    temp2.append(matrix[i][k][j])
                temp.append(temp2)
            output.append(temp)
        return output

    multiplied = get_self_multiplied_permutation(permutation_order)
    return _transpose(multiplied)


def permute_matrix_rowwise(matrix, main_permutation_order):
    lp = LimitedPermutation(main_permutation_order=main_permutation_order, input_list=[1, 2, 3, 4, 5, 6, 7])
    output = []
    for row in matrix:
        permutation_order = lp.__next__()
        output.append(permute(input_list=row, permutation_order=permutation_order))
    return output


def invert_matrix(matrix):
    return [x for x in zip(*matrix)]


def permute_matrix_columnwise(matrix, main_permutation_order):
    inverted_matrix = invert_matrix(matrix)
    inverted_matrix = permute_matrix_rowwise(inverted_matrix, main_permutation_order)
    return invert_matrix(inverted_matrix)


def permute_matrix(matrix, main_permutation_order):
    output = permute_matrix_rowwise(matrix, main_permutation_order)
    return permute_matrix_columnwise(output, main_permutation_order)


class LimitedPermutation(object):
    class Element(object):
        def __init__(self, value, order):
            self.value = value
            '''order of element in the original input_list'''
            self.order = order

    def __init__(self, input_list, main_permutation_order, multi=(1, 1), reading_direction='horizontal'):

        self._iterator = None
        self._element_generator = None
        self._multiplied_orders = None
        self._multi = None
        self._reading_direction = None

        self.input_list = input_list
        self.main_permutation_order = list(main_permutation_order)
        self.multi = multi
        self.reading_direction = reading_direction

    @staticmethod
    def get_next_multi(multi, input_length):
        m_1 = multi[0]
        m_2 = multi[1] + 1
        m_1, m_2 = ((m_1 - 1) % input_length) + 1, ((m_2 - 1) % input_length) + 1
        return m_1, m_2

    @property
    def reading_direction(self):
        return self._reading_direction

    @reading_direction.setter
    def reading_direction(self, val):
        if val not in ('horizontal', 'vertical'):
            raise ValueError()
        self._multiplied_orders = None
        self._reading_direction = val

    @property
    def multi(self):
        return self._multi

    @multi.setter
    def multi(self, val):
        m_1 = val[0]
        m_2 = val[1]
        input_length = len(self.input_list)
        m_1, m_2 = (((m_1 - 1) % input_length) + 1 + ((m_2 - 1) // input_length)) % input_length, (
                (m_2 - 1) % input_length) + 1
        if m_1 == 0:
            m_1 = len(self.input_list)
        self._multi = (m_1, m_2)

    @property
    def multiplied_orders(self):
        if self._multiplied_orders is None:
            if self.reading_direction == 'horizontal':
                multiplied = get_reordered_self_multiplied_permutation(self.main_permutation_order)
            else:
                multiplied = get_vertical_self_multiplied_permutation(self.main_permutation_order)

            self._multiplied_orders = [order for orders in multiplied for order in orders]
        return self._multiplied_orders

    @property
    def permutation_order(self):
        index = (self.multi[0] - 1) * len(self.input_list) + self.multi[1] - 1
        return self.multiplied_orders[index]

    @property
    def iterator(self):
        if self._iterator is None:
            self._iterator = cycle(self.multiplied_orders)
            if self.multi is None:
                raise Exception('multi must be set first')

            first_index = (self.multi[0] - 1) * len(self.main_permutation_order) + (self.multi[1] - 1)
            for i in range(first_index):
                self._iterator.__next__()

        return self._iterator

    def __next__(self):
        return self.iterator.__next__()

    @property
    def element_generator(self):
        if self._element_generator is None:
            def gen():
                next_permutations = self.__next__()
                while True:
                    for order in next_permutations:
                        yield self.Element(value=self.input_list[order - 1], order=order)

                    next_permutations = self.__next__()

            self._element_generator = gen()

        return self._element_generator

    def next_element(self):
        return self.element_generator.__next__()

    @property
    def next_multi(self):
        return self.get_next_multi(self.multi, len(self.input_list))



from musurgia.pdf.masterslave import Master, PositionMaster, MarginMaster
from musurgia.unittest import TestCase


class DummyMaster(Master):

    def get_slave_position(self, slave, position):
        pass

    def get_slave_margin(self, slave, margin):
        pass


class DummyPositionMaster(PositionMaster):
    def get_slave_position(self, slave, position):
        pass


class DummyMarginMaster(MarginMaster):
    def get_slave_margin(self, slave, margin):
        pass


class TestMaster(TestCase):
    def test_init(self):
        m = DummyMaster()
        self.assertTrue(isinstance(m, Master))

    def test_init_margin(self):
        m = DummyMarginMaster()
        self.assertTrue(isinstance(m, MarginMaster))

    def test_init_position(self):
        m = DummyPositionMaster()
        self.assertTrue(isinstance(m, PositionMaster))



from musurgia.pdf.masterslave import PositionMaster
from musurgia.pdf.text import TextLabel


class Labeled(PositionMaster):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._above_text_labels = []
        self._below_text_labels = []
        self._left_text_labels = []

    def add_text_label(self, label, **kwargs):
        if not isinstance(label, TextLabel):
            label = TextLabel(label, **kwargs)
        label.master = self
        if label.placement == 'above':
            self._above_text_labels.append(label)
        elif label.placement == 'below':
            self._below_text_labels.append(label)
        elif label.placement == 'left':
            self._left_text_labels.append(label)
        else:
            raise AttributeError()

        return label

    def add_label(self, label, **kwargs):
        return self.add_text_label(label, **kwargs)

    @property
    def above_text_labels(self):
        return self._above_text_labels

    @property
    def below_text_labels(self):
        return self._below_text_labels

    @property
    def left_text_labels(self):
        return self._left_text_labels

    @property
    def text_labels(self):
        return self.left_text_labels + self.above_text_labels + self.below_text_labels

    def draw_above_text_labels(self, pdf):
        if self.above_text_labels:
            with pdf.saved_state():
                translate_y = -self.get_above_text_labels_height() + self.above_text_labels[-1].get_text_height()
                pdf.translate(0, translate_y)
                for text_label in self.above_text_labels:
                    text_label.draw(pdf)
                    pdf.translate(0, text_label.get_height())

    def draw_below_text_labels(self, pdf):
        if self.below_text_labels:
            with pdf.saved_state():
                pdf.translate(self.relative_x, self.get_height())
                for text_label in self.below_text_labels:
                    pdf.translate(0, text_label.get_height())
                    text_label.draw(pdf)

    def draw_left_text_labels(self, pdf):
        if self.left_text_labels:
            with pdf.saved_state():
                pdf.translate(0, -self.get_left_text_labels_height() / 2)
                for text_label in self.left_text_labels:
                    pdf.translate(0, text_label.get_height())
                    with pdf.saved_state():
                        pdf.translate(-(text_label.get_width()), 0)
                        text_label.draw(pdf)

    def get_slave_position(self, slave, position):
        if position == 'x':
            return 0
        elif position == 'y':
            if slave.placement in ['l', 'left']:
                return self.get_height() / 2
            return 0
        else:
            raise AttributeError(position)

    def get_above_text_labels_height(self):
        return sum([tl.get_height() for tl in self.above_text_labels])

    def get_left_text_labels_height(self):
        return sum([tl.get_height() for tl in self.left_text_labels])



from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase


class TestRowContainer(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_get_relative_x2(self):
        self.assertFalse(True)

    def test_get_relative_y2(self):
        self.assertFalse(True)

    def test_get_height(self):
        self.assertFalse(True)

    def test_get_width(self):
        self.assertFalse(True)



import os
from unittest import TestCase
from quicktions import Fraction

from musurgia.fractaltree.fractaltree import FractalTree

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):

    def test_1(self):
        ft = FractalTree(proportions=(1, 2, 3), tree_permutation_order=(3, 1, 2), value=10)
        ft.add_layer()
        ft.add_layer()
        for node in ft.get_layer(1):
            node.reduce_children(condition=lambda node: node.fractal_order == 1)
        self.assertEqual([node.fractal_order for node in ft.traverse_leaves()], [2, 3, 3, 2, 2, 3])

    def test_2(self):
        ft = FractalTree(proportions=(1, 2, 3), tree_permutation_order=(3, 1, 2), value=10)
        ft.add_layer()
        ft.add_layer()
        for node in ft.get_layer(1):
            node.reduce_children(condition=lambda node: node.fractal_order == 1)
        actual = [node.value for node in ft.traverse_leaves()]
        expected = [Fraction(2, 1),
                    Fraction(3, 1),
                    Fraction(1, 1),
                    Fraction(2, 3),
                    Fraction(4, 3),
                    Fraction(2, 1)]
        self.assertEqual(actual, expected)

    def test_3(self):
        ft = FractalTree(proportions=[1, 2, 3, 4, 5, 6], tree_permutation_order=[4, 1, 5, 3, 6, 2], value=20,
                         multi=(4, 3))
        ft.add_layer()
        ft.reduce_children(lambda child: child.fractal_order not in [2, 3])
        actual = [node.value for node in ft.get_children()]
        expected = [12, 8]
        self.assertEqual(actual, expected)


from pathlib import Path

from musurgia.pdf.line import HorizontalLineSegment, VerticalLineSegment
from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase

path = Path(__file__)


class TestStraightLineLabels(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.hls = HorizontalLineSegment(length=20)
        self.hls.start_mark_line.show = False
        self.vls = VerticalLineSegment(length=20)
        self.vls.start_mark_line.show = False

    def test_horizontal_above(self):
        self.hls.straight_line.add_text_label('one above')
        self.hls.straight_line.add_text_label('two above')
        self.hls.straight_line.add_text_label('three above')
        with self.file_path(path, 'horizontal_above', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(20, 20)
            self.hls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_horizontal_below(self):
        self.hls.straight_line.add_text_label('one below', placement='below')
        self.hls.straight_line.add_text_label('two below', placement='below')
        self.hls.straight_line.add_text_label('three below', placement='below')
        with self.file_path(path, 'test_horizontal_below', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(20, 20)
            self.hls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_horizontal_left(self):
        self.hls.straight_line.add_text_label('one left', placement='left')
        self.hls.straight_line.add_text_label('two left', placement='left')
        self.hls.straight_line.add_text_label('three left', placement='left')
        with self.file_path(path, 'horizontal_left', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(20, 20)
            self.hls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_vertical_above(self):
        self.vls.straight_line.add_text_label('one above')
        self.vls.straight_line.add_text_label('two above')
        self.vls.straight_line.add_text_label('three above')
        with self.file_path(path, 'vertical_above', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(20, 20)
            self.vls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_vertical_above_with_relative_y(self):
        self.vls.relative_y = -10
        # print(self.vls.straight_line.relative_y)
        self.vls.straight_line.add_text_label('one above')
        self.vls.straight_line.add_text_label('two above')
        self.vls.straight_line.add_text_label('three above')
        with self.file_path(path, 'vertical_above_with_relative_y', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(20, 20)
            self.vls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_vertical_below(self):
        self.vls.straight_line.add_text_label('one below', placement='below')
        self.vls.straight_line.add_text_label('two below', placement='below')
        self.vls.straight_line.add_text_label('three below', placement='below')
        with self.file_path(path, 'test_vertical_below', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(20, 20)
            self.vls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_vertical_left(self):
        self.vls.straight_line.add_text_label('one left', placement='left')
        self.vls.straight_line.add_text_label('two left', placement='left')
        self.vls.straight_line.add_text_label('three left', placement='left')
        with self.file_path(path, 'vertical_left', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(20, 20)
            self.vls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_vertical_left_with_top_margin(self):
        self.vls.top_margin = 20
        self.vls.straight_line.add_text_label('one left', placement='left')
        self.vls.straight_line.add_text_label('two left', placement='left')
        self.vls.straight_line.add_text_label('three left', placement='left')
        with self.file_path(path, 'vertical_left_with_top_margin', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(20, 20)
            self.vls.draw(self.pdf)
            self.pdf.write(pdf_path)



from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from quicktions import Fraction

from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.basic_functions import dToX
from musurgia.chordfield.valuegenerator import ValueGenerator, ValueGeneratorException
from musurgia.quantize import get_quantized_values, find_best_quantized_values


class BreatheException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class ChordFieldException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class NoNextChordError(ChordFieldException):
    def __init__(self, *args):
        super().__init__(*args)


class LongEndingError(ChordFieldException):
    def __init__(self, delta, *args):
        msg = 'delta={}'.format(float(delta))
        super().__init__(msg, *args)


class ShortEndingError(ChordFieldException):
    def __init__(self, delta, *args):
        msg = 'delta={}'.format(float(delta))
        super().__init__(msg, *args)


class ParentSetQuarterDurationError(ChordFieldException):
    def __init__(self, *args):
        msg = 'parent\'s quarter_duration cannot be set'
        super().__init__(msg, *args)


class ParentSetEndingModesError(ChordFieldException):
    def __init__(self, *args):
        msg = 'parent\'s endings modes cannot be set'
        super().__init__(msg, *args)


class ChordField(object):
    def __init__(self, quarter_duration=None, duration_generator=None, midi_generator=None, chord_generator=None,
                 long_ending_mode=None, short_ending_mode=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._quarter_duration = None
        self._duration_generator = None
        self._midi_generator = None
        self._chord_generator = None
        self._chords = None
        self._long_ending_mode = None
        self._short_ending_mode = None
        self._children = None
        self._parent = None
        self._position = 0
        self._name = None

        self.quarter_duration = quarter_duration
        self.duration_generator = duration_generator
        self.midi_generator = midi_generator
        self.chord_generator = chord_generator
        self.long_ending_mode = long_ending_mode
        self.short_ending_mode = short_ending_mode

    def _get_value_generators(self):
        return [value_generator for value_generator in
                (self.chord_generator, self.duration_generator, self.midi_generator) if value_generator is not None]

    def _get_private_value_generators(self):
        return [value_generator for value_generator in
                (self._chord_generator, self._duration_generator, self._midi_generator) if value_generator is not None]

    def _set_up_value_generator(self, value_generator, value_mode=None):
        if not isinstance(value_generator, ValueGenerator):
            raise TypeError('value_generator must be of type ValueGenerator not {}'.format(
                type(value_generator)))
        value_generator.value_mode = value_mode
        try:
            value_generator.duration = self.quarter_duration
        except ValueGeneratorException:
            if self.quarter_duration is None:
                raise ValueGeneratorException()
            factor = Fraction(self.quarter_duration, value_generator.duration)
            for child in value_generator.children:
                child.duration *= factor

    @property
    def children(self):
        return self._children

    @property
    def parent(self):
        return self._parent

    def add_child(self, child):
        if self._quarter_duration:
            raise ParentSetQuarterDurationError()
        if self.long_ending_mode is not None or self.short_ending_mode is not None:
            raise ParentSetEndingModesError()
        if not isinstance(child, ChordField):
            raise TypeError()
        if self._children is None:
            self._children = []

        self._children.append(child)
        if child.duration_generator:
            if self.duration_generator is None:
                self.duration_generator = ValueGenerator()
                self.duration_generator.add_child(child.duration_generator)
            elif self.duration_generator.children:
                self.duration_generator.add_child(child.duration_generator)
            else:
                pass
        if child.midi_generator:
            if self.midi_generator is None:
                self.midi_generator = ValueGenerator()
                self.midi_generator.add_child(child.midi_generator)
            elif self.midi_generator.children:
                self.midi_generator.add_child(child.midi_generator)
            else:
                pass
        if child.chord_generator:
            if self.chord_generator is None:
                self.chord_generator = ValueGenerator()
                self.chord_generator.add_child(child.chord_generator)
            elif self.chord_generator.children:
                self.chord_generator.add_child(child.chord_generator)
            else:
                pass
        child._parent = self
        self._update_durations()
        return child

    def _update_durations(self):
        for value_generator in self._get_value_generators():
            try:
                value_generator.duration = self.quarter_duration
            except ValueGeneratorException:
                pass

    @property
    def quarter_duration(self):
        if self.children:
            children_quarter_durations = [child.quarter_duration for child in self.children]
            if None in children_quarter_durations:
                return None
            else:
                return sum(children_quarter_durations)
        return self._quarter_duration

    @quarter_duration.setter
    def quarter_duration(self, value):
        if value is not None:
            if self.children:
                raise ParentSetQuarterDurationError()
            else:
                self._quarter_duration = value
                self._update_durations()
                if self.parent:
                    self.parent._update_durations()

    @property
    def duration_generator(self):
        if not self._duration_generator and self.parent and self.parent.duration_generator:
            return self.parent.duration_generator
        return self._duration_generator

    @duration_generator.setter
    def duration_generator(self, value):
        self._duration_generator = value
        if value:
            self._set_up_value_generator(self.duration_generator, 'duration')

    @property
    def midi_generator(self):
        if not self._midi_generator and self.parent and self.parent.midi_generator:
            return self.parent.midi_generator
        return self._midi_generator

    @midi_generator.setter
    def midi_generator(self, value):
        self._midi_generator = value
        if value:
            self._set_up_value_generator(self.midi_generator, 'midi')

    @property
    def chord_generator(self):
        if not self._chord_generator and self.parent:
            return self.parent.chord_generator
        return self._chord_generator

    @chord_generator.setter
    def chord_generator(self, value):
        self._chord_generator = value
        if value:
            self._set_up_value_generator(self.chord_generator, 'chord')

    @property
    def position(self):
        return self._position

    @property
    def position_in_parent(self):
        index = self.parent.children.index(self)
        if index == 0:
            return 0
        return sum([child.quarter_duration for child in self.parent.children[:index]])

    @property
    def long_ending_mode(self):
        return self._long_ending_mode

    @long_ending_mode.setter
    def long_ending_mode(self, val):
        if val is not None and self.children is not None:
            raise ParentSetEndingModesError()
        """
        :param val: can be None, 'self_extend', 'cut', 'omit', 'omit_and_add_rest', 'omit_and_stretch'
        for dealing with last chord, if it is too long and ends after self.quarter_duration
        None: raises Error
        self_extend: self.quarter_duration will be prolonged.
        cut: last chord will be cut short.
        omit: last chord will be omitted and self.quarter_duration cut short.
        omit_and_add_rest: last chord will be omitted and rests will be added.
        omit_and_stretch: last chord will be omitted and the new last chord  will be extended.
        """
        permitted = [None, 'self_extend', 'cut', 'omit', 'omit_and_add_rest', 'omit_and_stretch']
        if val not in permitted:
            raise ValueError('long_ending_mode.value {} must be in {}'.format(val, permitted))
        self._long_ending_mode = val

    @property
    def short_ending_mode(self):
        return self._short_ending_mode

    @short_ending_mode.setter
    def short_ending_mode(self, val):
        if val is not None and self.children is not None:
            raise ParentSetEndingModesError()
        """
        :param val: can be None, 'self_shrink', 'add_rest', 'stretch'
        for dealing with last chord, if it is too long and ends after self.quarter_duration
        None: raises Error
        self_shrink: self.quarter_duration will be shortened.
        omit: rests will be added.
        stretch: last chord will be prolonged.
        """
        permitted = [None, 'self_shrink', 'add_rest', 'stretch']
        if val not in permitted:
            raise ValueError('short_ending_mode.value {} must be in {}'.format(val, permitted))
        self._short_ending_mode = val

    @property
    def chords(self):
        if self.children:
            self._chords = [chord for child in self.children for chord in child.chords]
        else:
            list(self)
        return self._chords

    @property
    def simple_format(self):
        sf = SimpleFormat()
        if self.chords:
            for chord in self.chords:
                sf.add_chord(chord)
        return sf

    def _get_next_duration(self):
        if self.duration_generator:
            next_duration = self.duration_generator.__next__()
            self._position += next_duration
            for generator in self._get_value_generators():
                if generator != self.duration_generator:
                    generator._position += next_duration
            return next_duration
        else:
            return None

    def _get_next_midi(self):
        if self.midi_generator:
            # self.midi_generator._position = self.position
            return self.midi_generator.__next__()
        else:
            return None

    def _get_next_chord(self):
        next_chord = None
        next_midi = self._get_next_midi()
        next_duration = self._get_next_duration()

        if self.chord_generator:
            next_chord = self.chord_generator.__next__()

        if next_chord:
            if next_duration is not None:
                next_chord.quarter_duration = next_duration
            if next_midi is not None:
                next_chord.midis = next_midi
        else:
            if next_duration is None:
                raise NoNextChordError('next_duration is None!')
            if next_midi is None:
                raise NoNextChordError('next_midi is None!')
            next_chord = TreeChord(quarter_duration=next_duration, midis=next_midi)
        if self._chords is None:
            self._chords = []
        self._chords.append(next_chord)
        return next_chord

    def _check_quarter_duration(self):
        if self._chords:
            delta = sum([chord.quarter_duration for chord in self._chords]) - self.quarter_duration
            if delta > 0:
                if self.long_ending_mode == 'self_extend':
                    if self.parent:
                        try:
                            next_child = self.parent.children[self.parent.children.index(self) + 1]
                            next_child._position = delta
                        except IndexError:
                            pass
                    try:
                        self.quarter_duration += delta
                    except ParentSetQuarterDurationError:
                        self.children[-1].quarter_duration += delta
                elif self.long_ending_mode == 'cut':
                    self._chords[-1].quarter_duration -= delta
                elif self.long_ending_mode in ['omit', 'omit_and_add_rest', 'omit_and_stretch']:
                    self._chords.pop()
                    new_delta = self.quarter_duration - sum([chord.quarter_duration for chord in self._chords])
                    if self.long_ending_mode == 'omit_and_add_rest':
                        new_chord = TreeChord(midis=0, quarter_duration=new_delta)
                        new_chord.zero_mode = 'remove'
                        self._chords.append(new_chord)
                    elif self.long_ending_mode == 'omit_and_stretch':
                        self._chords[-1].quarter_duration += new_delta
                    else:
                        self.quarter_duration -= new_delta
                else:
                    raise LongEndingError(delta)

            elif delta < 0:
                if self.short_ending_mode == 'self_shrink':
                    self.quarter_duration += delta
                elif self.short_ending_mode == 'add_rest':
                    new_chord = TreeChord(midis=0, quarter_duration=-delta)
                    new_chord.zero_mode = 'remove'
                    self._chords.append(new_chord)
                elif self.short_ending_mode == 'stretch':
                    self._chords[-1].quarter_duration -= delta
                else:
                    raise ShortEndingError(delta)
            else:
                pass

    def __iter__(self):
        return self

    def __next__(self):
        try:
            next_chord = self._get_next_chord()
            return next_chord
        except StopIteration:
            self._check_quarter_duration()
            raise StopIteration()

    def __deepcopy__(self, memodict={}):
        if self.__class__ == Breathe:
            copied = self.__class__(proportions=self.proportions, breakpoints=self.breakpoints,
                                    quarter_duration=self.quarter_duration)
        else:
            copied = self.__class__(quarter_duration=self.quarter_duration)

        copied.long_ending_mode = self.long_ending_mode
        copied.short_ending_mode = self.short_ending_mode
        for generator in self._get_private_value_generators():
            if generator.value_mode == 'duration':
                copied._duration_generator = generator.__deepcopy__()
            elif generator.value_mode == 'midi':
                copied._midi_generator = generator.__deepcopy__()
            elif generator.value_mode == 'chord':
                copied._chord_generator = generator.__deepcopy__()
            else:
                raise NotImplementedError()
        if self.children:
            copied._children = []
            for child in self.children:
                copied_child = child.__deepcopy__()
                copied_child._parent = copied
                copied._children.append(copied_child)
        return copied


class Breathe(ChordField):
    def __init__(self, proportions, breakpoints=None, quarter_duration=None, quantize=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._proportions = None
        self._breakpoints = None
        self._quantize = None
        self.quantize = quantize
        self.proportions = proportions
        self.breakpoints = breakpoints
        self.quarter_duration = quarter_duration

    @property
    def quarter_durations(self):
        return [self.repose_1.quarter_duration, self.inspiration.quarter_duration, self.climax.quarter_duration,
                self.expiration.quarter_duration, self.repose_2.quarter_duration]

    def _generate_children(self, quarter_duration):
        names = ['repose_1', 'inspiration', 'climax', 'expiration', 'repose_2']
        fields = []
        child_quarter_durations = [Fraction(proportion * quarter_duration, sum(self.proportions)) for proportion in
                                   self.proportions]
        if self.quantize:
            child_quarter_durations = get_quantized_values(child_quarter_durations, grid_size=self.quantize)

        for index, child_quarter_duration in enumerate(child_quarter_durations):
            field = ChordField(quarter_duration=child_quarter_duration)
            field.name = names[index]
            fields.append(field)

        if not self.breakpoints:
            self.breakpoints = [0.25, 1, 0.25]

        for i in range(5):
            if i == 0:
                a1, an = self.breakpoints[0], self.breakpoints[0]
            elif i == 1:
                a1, an = self.breakpoints[0], self.breakpoints[1]
            elif i == 2:
                a1, an = self.breakpoints[1], self.breakpoints[1]
            elif i == 3:
                a1, an = self.breakpoints[1], self.breakpoints[2]
            else:
                a1, an = self.breakpoints[2], self.breakpoints[2]
            fields[i].duration_generator = ValueGenerator(ArithmeticProgression(a1=a1, an=an, correct_s=True))
        for field in fields:
            if field.quarter_duration:
                generator = field.duration_generator.generator
                if generator.n == 0:
                    raise BreatheException(
                        'field {}: ArithmeticProgression: a1={} s={} n={}'.format(field.name, generator.a1, generator.s,
                                                                                  generator.n))
                self.add_child(field)

    @ChordField.quarter_duration.setter
    def quarter_duration(self, value):
        if value is not None:
            self._generate_children(value)

    @property
    def proportions(self):
        return self._proportions

    @proportions.setter
    def proportions(self, values):
        if not self._proportions:
            if len(values) != 5:
                ValueError('quarter_durations must have 5 elements.')
            self._proportions = values
        else:
            raise BreatheException('proportions can only be set during initialisation')

    @property
    def breakpoints(self):
        return self._breakpoints

    @breakpoints.setter
    def breakpoints(self, values):
        if not self._breakpoints:
            if len(values) != 3:
                ValueError('quarter_durations must have 3 elements.')
            self._breakpoints = values
        else:
            raise BreatheException('breakpoints can only be set during initialisation')

    @property
    def repose_1(self):
        return self.children[0]

    @property
    def inspiration(self):
        return self.children[1]

    @property
    def climax(self):
        return self.children[2]

    @property
    def expiration(self):
        return self.children[3]

    @property
    def repose_2(self):
        return self.children[4]


class ProportionalChordField(ChordField):
    def __init__(self, quarter_duration, proportions=None, direction='up', min_midi=60, max_midi=72,
                 duration_units=None, midi_grid=1,
                 show_last_note=False,
                 *args, **kwargs):
        super().__init__(quarter_duration=quarter_duration, *args, **kwargs)
        self._proportions = None
        self._min_midi = None
        self._max_midi = None
        self._units = None
        self._midi_grid = None
        self._show_last_note = None
        self._midi_values = None
        self._quarter_durations = None
        self._direction = None

        self.min_midi = min_midi
        self.max_midi = max_midi
        self.direction = direction
        self.duration_units = duration_units
        self.midi_grid = midi_grid
        self.proportions = proportions
        self.show_last_note = show_last_note

    def _calculate_quarter_durations(self):
        if self.proportions:
            quarter_durations = [
                (Fraction(Fraction(p), Fraction(sum(self.proportions)))) * Fraction(self.quarter_duration) for p in
                self.proportions]

            if self.duration_units:
                quarter_durations = find_best_quantized_values(values=quarter_durations, units=self.duration_units,
                                                               check_sum=True)
            if self.direction == 'down':
                quarter_durations = quarter_durations[::-1]
            self._quarter_durations = quarter_durations
        if self.quarter_durations:
            self._duration_generator = ValueGenerator(iter(self.quarter_durations), duration=self.quarter_duration,
                                                      value_mode='duration')
        else:
            self._duration_generator = None

    def _calculate_midi_values(self):
        if self.proportions:
            midi_intervals = [(p / sum(self.proportions)) * (self.max_midi - self.min_midi) for p in self.proportions]
            if self.direction == 'up':
                midi_values = get_quantized_values(dToX(midi_intervals, first_element=self.min_midi),
                                                   grid_size=self.midi_grid)
            else:
                midi_intervals = reversed(midi_intervals)
                midi_intervals = [-1 * interval for interval in midi_intervals]
                midi_values = get_quantized_values(dToX(midi_intervals, first_element=self.max_midi),
                                                   grid_size=self.midi_grid)
            midi_values = [float(midi) for midi in midi_values]

            self._midi_values = midi_values
        if self.midi_values:
            self._midi_generator = ValueGenerator(iter(self.midi_values), value_mode='midi',
                                                  duration=self.quarter_duration)
        else:
            self._midi_generator = None

    @property
    def chords(self):
        if not self.proportions:
            raise AttributeError('set proportions first')
        return super().chords

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, val):
        permitted = ['up', 'down']
        if val not in permitted:
            raise ValueError(f'direction.value {val} must be in {permitted}')
        self._direction = val
        self._calculate_midi_values()

    @property
    def duration_units(self):
        return self._duration_units

    @duration_units.setter
    def duration_units(self, val):
        self._duration_units = val
        self._calculate_quarter_durations()

    @property
    def midi_grid(self):
        return self._midi_grid

    @midi_grid.setter
    def midi_grid(self, val):
        self._midi_grid = val
        self._calculate_midi_values()

    @property
    def midi_values(self):
        return self._midi_values

    @property
    def max_midi(self):
        return self._max_midi

    @max_midi.setter
    def max_midi(self, value):
        if self.min_midi and value < self.min_midi:
            raise AttributeError(f'max_midi cannot be smaller than min_midi{self.min_midi}')
        self._max_midi = value
        self._calculate_midi_values()

    @property
    def min_midi(self):
        return self._min_midi

    @min_midi.setter
    def min_midi(self, value):
        if self.max_midi and value > self.max_midi:
            raise AttributeError(f'min_midi cannot be bigger than max_midi{self.max_midi}')
        self._min_midi = value
        self._calculate_midi_values()

    @property
    def proportions(self):
        return self._proportions

    @proportions.setter
    def proportions(self, val):
        self._proportions = val
        self._calculate_quarter_durations()
        self._calculate_midi_values()

    @property
    def quarter_durations(self):
        return self._quarter_durations

    @property
    def show_last_note(self):
        return self._show_last_note

    @show_last_note.setter
    def show_last_note(self, val):
        if not isinstance(val, bool):
            raise TypeError(f"show_last_note.value must be of type bool not{type(val)}")
        self._show_last_note = val

    @property
    def simple_format(self):
        output = super().simple_format
        if self.show_last_note:
            if self.direction == 'up':
                grace_chord = TreeChord(midis=self.max_midi)
            else:
                grace_chord = TreeChord(midis=self.min_midi)
            output.chords[-1].add_grace_chords(grace_chord, mode='post')
        return output


class HalfWave(ProportionalChordField):
    pass


class Wave(ChordField):
    def __init__(self, quarter_duration, field_proportions, show_last_note=False, direction='up', *args, **kwargs):
        super().__init__(quarter_duration=quarter_duration, *args, **kwargs)
        self._quarter_duration = None
        self._field_proportions = None
        self._show_last_note = None
        self._direction = None
        self.direction = direction
        self.show_last_note = show_last_note

        self._rising_field = HalfWave(
            quarter_duration=quarter_duration * (field_proportions[0] / sum(field_proportions)), direction='up')
        self._falling_field = HalfWave(
            quarter_duration=quarter_duration * (field_proportions[1] / sum(field_proportions)), direction='down')
        if self.direction == 'up':
            self.add_child(self._rising_field)
            self.add_child(self._falling_field)
        else:
            self.add_child(self._falling_field)
            self.add_child(self._rising_field)

    @property
    def rising_field(self):
        return self._rising_field

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, val):
        if self._direction:
            raise AttributeError('direction can only be set by initialization')
        permitted = ['up', 'down']
        if val not in permitted:
            raise ValueError(f'direction.value {val} must be in {permitted}')
        self._direction = val

    @property
    def falling_field(self):
        return self._falling_field

    @property
    def simple_format(self):
        output = super().simple_format
        if self.show_last_note:
            if self.direction == 'up':
                grace_chord = TreeChord(midis=self.falling_field.min_midi)
            else:
                grace_chord = TreeChord(midis=self.rising_field.max_midi)
            output.chords[-1].add_grace_chords(grace_chord, mode='post')
        return output

    @property
    def show_last_note(self):
        return self._show_last_note

    @show_last_note.setter
    def show_last_note(self, val):
        if not isinstance(val, bool):
            raise TypeError(f"show_last_note.value must be of type bool not{type(val)}")
        self._show_last_note = val


from musurgia.fractaltree.fractaltree import FractalTree
from musurgia.tree import Tree
from musurgia.unittest import TestCase


class TestFtSplit(TestCase):
    def test_two(self):
        t = FractalTree(value=10)
        t.add_layer()
        child = t.get_children()[0]
        s = child.split(1, 1)
        actual = [x.value for x in s]
        expected = [child.value/2, child.value/2]
        self.assertEqual(expected, actual)

    def test_three(self):
        t = FractalTree(value=10)
        t.add_layer()
        child = t.get_children()[0]
        s = child.split(1, 1, 1)
        actual = [x.value for x in s]
        expected = [child.value/3, child.value/3, child.value/3]
        self.assertEqual(expected, actual)

    def test_child_is_not_leaf(self):
        t = FractalTree(value=10)
        t.add_layer()
        child = t.get_children()[0]
        child.split(1, 1, 1)
        self.assertFalse(child.is_leaf)

    def test_split_is_leaf(self):
        t = FractalTree(value=10)
        t.add_layer()
        child = t.get_children()[0]
        s = child.split(1, 1, 1)
        actual = [x.is_leaf for x in s]
        expected = [True, True, True]
        self.assertEqual(expected, actual)



import os
from itertools import cycle
from math import ceil

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.random import Random
from musurgia.unittest import TestCase
from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.chordfield.chordfield import ChordField, ShortEndingError, LongEndingError
from musurgia.chordfield.valuegenerator import ValueGenerator
from musurgia.interpolation import Interpolation

path = str(os.path.abspath(__file__).split('.')[0])


def duration_generator(first_duration=1, delta=0.5):
    current_duration = first_duration
    while True:
        yield current_duration
        current_duration += delta


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        # chord_generator with matching quarter_duration
        field = ChordField(quarter_duration=11,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2, 1, 2, 3, 4, 5]).chords)))
        self.score.set_time_signatures(quarter_durations=field.quarter_duration)
        field.simple_format.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_1.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_2(self):
        # chord_generator with too long quarter_duration:  mode : None
        field = ChordField(quarter_duration=10,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2, 1, 2, 3, 4, 5]).chords)))
        with self.assertRaises(LongEndingError):
            list(field)

    def test_3(self):
        # chord_generator with too long quarter_duration:  mode : self_extend
        field = ChordField(quarter_duration=10,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2, 1, 2, 3, 4, 5]).chords)),
                           long_ending_mode='self_extend')
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=field.quarter_duration)
        sf.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_3.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_4(self):
        # chord_generator with too long quarter_duration:  mode : cut
        field = ChordField(quarter_duration=10,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2, 1, 2, 3, 4, 5]).chords)),
                           long_ending_mode='cut')
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=field.quarter_duration)
        sf.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_4.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_5(self):
        # chord_generator with too long quarter_duration:  mode : omit
        field = ChordField(quarter_duration=10,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2, 1, 2, 3, 4, 5]).chords)),
                           long_ending_mode='omit')
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=field.quarter_duration)
        sf.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_5.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_6(self):
        # chord_generator with too long quarter_duration:  mode : omit
        field = ChordField(quarter_duration=10,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2, 1, 2, 3, 4, 5]).chords)),
                           long_ending_mode='omit_and_add_rest')
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=field.quarter_duration)
        sf.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_6.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_7(self):
        # chord_generator with too long quarter_duration:  mode : omit
        field = ChordField(quarter_duration=10,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2, 1, 2, 3, 4, 5]).chords)),
                           long_ending_mode='omit_and_stretch')
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=field.quarter_duration)
        sf.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_7.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_8(self):
        # chord_generator with too short quarter_duration:  mode : None
        field = ChordField(quarter_duration=10,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2]).chords)))
        with self.assertRaises(ShortEndingError):
            list(field)

    def test_9(self):
        # chord_generator with too short quarter_duration:  mode : self_shrink
        field = ChordField(quarter_duration=10,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2]).chords)),
                           short_ending_mode='self_shrink')
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=field.quarter_duration)
        sf.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_9.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_10(self):
        # chord_generator with too short quarter_duration:  mode : stretch
        field = ChordField(quarter_duration=10,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2]).chords)),
                           short_ending_mode='stretch')
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=field.quarter_duration)
        sf.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_10.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_11(self):
        # chord_generator with too short quarter_duration:  mode : rest
        field = ChordField(quarter_duration=10,
                           chord_generator=ValueGenerator(
                                iter(SimpleFormat(quarter_durations=[3, 2]).chords)),
                           short_ending_mode='add_rest')
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=field.quarter_duration)
        sf.to_stream_voice().add_to_score(self.score)

        xml_path = path + '_test_11.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_12(self):
        field = ChordField(quarter_duration=10,
                           duration_generator=ValueGenerator(
                                cycle([1])),
                           midi_generator=ValueGenerator(
                                cycle([71]))
                           )
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=field.quarter_duration)
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_12.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_13(self):
        field = ChordField(quarter_duration=10,
                           duration_generator=ValueGenerator(
                                Random(pool=[0.2, 0.4, 0.8, 1.2, 1.6, 2], periodicity=3, seed=20)),
                           midi_generator=ValueGenerator(
                                cycle([71])),
                           long_ending_mode='self_extend'
                           )
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=ceil(field.quarter_duration))
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_13.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_14(self):
        field = ChordField(quarter_duration=10,
                           duration_generator=ValueGenerator(duration_generator(first_duration=1, delta=0.2)
                                                              ),
                           midi_generator=ValueGenerator(cycle([71])),
                           long_ending_mode='omit_and_add_rest'
                           )
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=ceil(field.quarter_duration))
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_14.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_15(self):
        field = ChordField(quarter_duration=10,
                           duration_generator=ValueGenerator(
                                Random(pool=[0.2, 0.4, 0.8, 1.2, 1.6, 2], periodicity=3, seed=20)),
                           midi_generator=ValueGenerator(
                                Interpolation(start=84, end=60, key=lambda x: int(x))),
                           long_ending_mode='self_extend'
                           )
        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=ceil(field.quarter_duration))
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_15.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_16(self):
        field = ChordField(quarter_duration=10,
                           duration_generator=ValueGenerator(ArithmeticProgression(a1=0.2, an=2, correct_s=True)),
                           midi_generator=ValueGenerator(Interpolation(start=84, end=60,
                                                                        key=lambda midi: round(midi * 2) / 2)),
                           short_ending_mode=None)

        sf = field.simple_format
        self.score.set_time_signatures(quarter_durations=ceil(field.quarter_duration))
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_16.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_17(self):
        parent_field = ChordField()
        child_field_1 = ChordField(quarter_duration=5,
                                   duration_generator=ValueGenerator(cycle([1])),
                                   midi_generator=ValueGenerator(cycle([60]))
                                   )
        child_field_2 = ChordField(quarter_duration=10,
                                   duration_generator=ValueGenerator(cycle([2])),
                                   midi_generator=ValueGenerator(cycle([61]))
                                   )
        parent_field.add_child(child_field_1)
        parent_field.add_child(child_field_2)

        sf = parent_field.simple_format
        self.score.set_time_signatures(quarter_durations=ceil(parent_field.quarter_duration))
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_17.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)



from pathlib import Path

from musurgia.pdf.line import VerticalSegmentedLine
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.text import Text
from musurgia.unittest import TestCase

path = Path(__file__)


class TestAddPage(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_absolute_y(self):
        self.pdf.translate(10, 10)
        self.pdf.add_page()
        actual = self.pdf.absolute_positions
        expected = {1: [10, 10], 2: [0, 0]}
        self.assertEqual(expected, actual)

    def test_draw(self):
        self.pdf.translate_page_margins()
        self.pdf.draw_ruler(mode='h')
        self.pdf.draw_ruler(mode='v')
        self.pdf.translate(10, 10)
        vsl = VerticalSegmentedLine(lengths=[10, 20, 30, 40])
        vsl.draw(self.pdf)
        self.pdf.add_page()
        self.pdf.translate_page_margins()
        self.pdf.draw_ruler(mode='h')
        self.pdf.draw_ruler(mode='v')
        self.pdf.translate(10, 10)
        # print(self.pdf.absolute_positions)
        vsl.draw(self.pdf)
        with self.file_path(path, 'draw', 'pdf') as pdf_path:
            self.pdf.write(pdf_path)

    def test_page_number(self):
        for i in range(3):
            self.pdf.add_page()
        self.pdf.draw_page_numbers(v_position='center', h_position='bottom')

        with self.file_path(path, 'page_number', 'pdf') as pdf_path:
            self.pdf.write(pdf_path)

    def test_draw_in_reversed_order(self):
        self.pdf.translate(30, 30)
        self.pdf.add_page()
        self.pdf.translate(80, 80)
        self.pdf.add_page()
        self.pdf.page = 2
        text = Text('second page')
        text.draw(self.pdf)
        self.pdf.page = 1
        text = Text('first page')
        text.draw(self.pdf)

        for page in self.pdf.pages:
            self.pdf.page = page
            self.pdf.reset_position()
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')

        self.pdf.draw_page_numbers(v_position='center', h_position='bottom')

        with self.file_path(path, 'draw_in_reversed_order', 'pdf') as pdf_path:
            self.pdf.write(pdf_path)



from pathlib import Path

from musurgia.pdf.line import VerticalLineSegment
from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase

path = Path(__file__)


class TestVerticalLineSegment(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.vls = VerticalLineSegment(length=10)

    def test_get_relative_x2(self):
        actual = self.vls.get_relative_x2()
        expected = self.vls.get_width()
        self.assertEqual(expected, actual)

    def test_start_mark_line_relative_x(self):
        actual = self.vls.start_mark_line.relative_x
        expected = -1.5
        self.assertEqual(expected, actual)

    def test_start_mark_line_relative_y(self):
        actual = self.vls.start_mark_line.relative_y
        expected = self.vls.relative_y
        self.assertEqual(expected, actual)

    def test_end_mark_line_relative_y(self):
        actual = self.vls.end_mark_line.relative_y
        expected = self.vls.relative_y + self.vls.length
        self.assertEqual(expected, actual)

    def test_end_mark_line_relative_x(self):
        actual = self.vls.end_mark_line.relative_x
        expected = -1.5
        self.assertEqual(expected, actual)

    def test_start_mark_line_top_margin(self):
        actual = self.vls.start_mark_line.top_margin
        expected = 0
        self.assertEqual(expected, actual)

    def test_get_width(self):
        actual = self.vls.get_width()
        expected = 3
        self.assertEqual(expected, actual)

    def test_draw(self):
        with self.file_path(parent_path=path, name='draw', extension='pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            self.vls.end_mark_line.show = True
            self.vls.draw(self.pdf)
            self.pdf.write(pdf_path)



import os

from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])

fm = FractalMusic(tempo=72, quarter_duration=18, tree_permutation_order=(2, 6, 4, 1, 3, 7, 5),
                  proportions=(1, 2, 3, 4, 5, 6, 7), multi=(1, 3), reading_direction='vertical')
fm.midi_generator.set_directions(-1, -1, -1, -1, -1, -1, -1)
fm.midi_generator.midi_range = [60, 72]
fm.generate_children(number_of_children=3)
for child in fm.get_children():
    child.add_gliss()

xml_path = path + '.xml'

score = fm.get_score(layer_number=fm.number_of_layers)
score.max_division = 5
score.write(xml_path)



from musurgia.pdf.newdrawobject import DrawObject


class Scene(DrawObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._draw_objects = []

    @property
    def draw_objects(self):
        return self._draw_objects

    def add_draw_object(self, draw_object):
        if not isinstance(draw_object, DrawObject):
            raise TypeError()
        self._draw_objects.append(draw_object)
        return draw_object

    def get_relative_x2(self):
        if self.draw_objects:
            return max([do.get_relative_x2() + do.right_margin + do.left_margin for do in self.draw_objects])
        else:
            return None

    def get_relative_y2(self):
        if self.draw_objects:
            return sum([do.get_relative_y2() + do.bottom_margin + do.bottom_margin for do in
                        self.draw_objects]) + self.relative_y
        else:
            return None

    def draw(self, pdf):
        # with pdf.saved_state():
        pdf.translate(self.relative_x, -self.relative_y)
        pdf.translate(self.left_margin, self.top_margin)
        for draw_object in self._draw_objects:
            draw_object.draw(pdf)
        pdf.translate(self.right_margin, self.bottom_margin)



from pathlib import Path

from musurgia.pdf.line import HorizontalLineSegment, HorizontalSegmentedLine, VerticalSegmentedLine
from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase

path = Path(__file__)


class TestMarkLineLabels(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.ls = HorizontalLineSegment(length=20)

    def test_draw_above(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label above')
        ml.add_text_label('second text label above')
        ml.add_text_label('third  text label above')
        self.ls.relative_x = 10
        self.ls.relative_y = 20
        with self.file_path(path, 'draw_above', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')

            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_one_above(self):
        ml = self.ls.start_mark_line
        ml.length = 20
        ml.add_text_label('first text label above')

        with self.file_path(path, 'draw_one_above', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            # self.pdf.draw_ruler('h')
            # self.pdf.draw_ruler('v')
            self.pdf.translate(30, 30)
            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_below(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label below', placement='below')
        ml.add_text_label('second text label below', placement='below')
        self.ls.relative_x = 10
        self.ls.relative_y = 20
        with self.file_path(path, 'draw_below', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_left(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label left', placement='left')
        ml.add_text_label('second text label left left left', placement='left')
        ml.add_text_label('third text label left left left', placement='left')
        ml.left_text_labels[1].font.size = 8
        self.ls.relative_x = 40
        self.ls.relative_y = 10
        with self.file_path(path, 'draw_left', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_different_sizes(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label above', font_size=7)
        ml.add_text_label('second text label above', font_size=8)
        ml.add_text_label('third text label above', font_size=9)
        with self.file_path(path, 'different_sizes', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 20)
            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_font_size_8(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label above', font_size=8, bottom_margin=2)
        ml.add_text_label('second text label above', font_size=8, bottom_margin=4)
        ml.add_text_label('third text label above', font_size=8)
        with self.file_path(path, 'font_size_8', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 20)
            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_below_with_different_mark_line_lengths(self):
        hsl = HorizontalSegmentedLine(lengths=[10, 15, 20])
        hsl.segments[0].start_mark_line.length = 6
        hsl.segments[0].start_mark_line.add_label(1, placement='below', font_size=8)
        hsl.segments[1].start_mark_line.add_label(2, placement='below', font_size=8)
        hsl.segments[2].start_mark_line.add_label(3, placement='below', font_size=8)
        with self.file_path(path, 'below_with_different_mark_line_lengths', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            hsl.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_vertical_with_left_text_labels(self):
        vsl = VerticalSegmentedLine(lengths=[10, 15, 20])
        vsl.segments[0].start_mark_line.length = 6
        vsl.segments[0].start_mark_line.add_label(1, placement='left', font_size=8)
        vsl.segments[1].start_mark_line.add_label(2, placement='left', font_size=8)
        vsl.segments[2].start_mark_line.add_label(3, placement='left', font_size=8)
        with self.file_path(path, 'vertical_with_left_text_labels', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            vsl.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_above_with_different_bottom_margins(self):
        ml = self.ls.start_mark_line
        ml.add_text_label('first text label above', bottom_margin=2)
        ml.add_text_label('second text label above', bottom_margin=4)
        ml.add_text_label('third  text label above', bottom_margin=15)
        self.ls.relative_x = 10
        self.ls.relative_y = 40
        with self.file_path(path, 'draw_above_with_different_bottom_margins', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')

            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_left_position(self):
        ml = self.ls.start_mark_line
        # print(ml.get_relative_position())
        # print(ml.get_height())
        left_l = ml.add_text_label('left one', placement='left')
        # print(left_l.get_relative_position())
        left_l = ml.add_text_label('left two', placement='left')
        left_l = ml.add_text_label('left three', placement='left')
        left_l = ml.add_text_label('left four', placement='left')

        with self.file_path(path, 'left_position', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(30, 30)
            self.ls.draw(self.pdf)
            self.pdf.write(pdf_path)



from pathlib import Path

from musurgia.pdf.line import MarkLine
from musurgia.pdf.masterslave import Master
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.text import Text
from musurgia.unittest import TestCase

path = Path(__file__)


class DummyMaster(Master):
    def get_slave_margin(self, slave, margin):
        return 10

    def get_slave_position(self, slave, position):
        return 20


class TestMarkLine(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.master = DummyMaster()
        self.ml = MarkLine(mode='v', placement='start', name='mark_line_test', master=self.master)

    def test_relative_x(self):
        actual = self.ml.relative_x
        expected = self.master.get_slave_position(self.ml, 'x')
        self.assertEqual(expected, actual)

    def test_get_relative_x2(self):
        actual = self.ml.get_relative_x2()
        expected = self.ml.relative_x
        self.assertEqual(expected, actual)

    def test_get_relative_y2(self):
        actual = self.ml.get_relative_y2()
        expected = self.ml.relative_y + self.ml.length
        self.assertEqual(expected, actual)

    def test_get_height(self):
        actual = self.ml.get_height()
        expected = self.ml.top_margin + self.ml.length + self.ml.bottom_margin
        self.assertEqual(expected, actual)

    def test_draw(self):
        with self.file_path(path, 'draw', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.ml.draw(self.pdf)
            self.pdf.write(pdf_path)

    def test_draw_multiple(self):
        self.ml.master.get_slave_position = lambda slave, position: 0
        self.ml.master.get_slave_margin = lambda slave, margin: 0
        with self.file_path(path, 'draw_multiple', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(20, 20)
            self.ml.draw(self.pdf)
            self.ml.master.get_slave_margin = lambda slave, margin: 5 if margin in ['l', 'left'] else 0
            self.ml.draw(self.pdf)
            self.pdf.write(pdf_path)



import os
from itertools import cycle

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.chordfield.chordfield import ChordField
from musurgia.chordfield.valuegenerator import ValueGenerator

path = str(os.path.abspath(__file__).split('.')[0])
score = TreeScoreTimewise()
parent_chord_field = ChordField(duration_generator=ValueGenerator(ArithmeticProgression(n=5, a1=0.5, correct_s=True)))
# parent_chord_field = ChordField(duration_generator=ValueGenerator(iter([0.5, 1, 1.5])))
quarter_durations = [3, 2, 3]
for i in range(len(quarter_durations)):
    quarter_duration = quarter_durations[i]
    midi = 60 + i
    parent_chord_field.add_child(
        ChordField(midi_generator=ValueGenerator(cycle([midi])), long_ending_mode='self_extend',
                   short_ending_mode='self_shrink', quarter_duration=quarter_duration))
###
print(list(parent_chord_field.duration_generator.__deepcopy__()))
simple_format = SimpleFormat()
for child in parent_chord_field.children:
    simple_format.extend(child.simple_format)
simple_format.to_stream_voice().add_to_score(score=score, part_number=3)
print(simple_format.quarter_duration)
print(float(simple_format.quarter_duration))
###

xml_path = path + '_test_2.xml'
score.write(xml_path)



from abc import ABC, abstractmethod
from math import ceil

from musurgia.pdf.margined import Margined
from musurgia.pdf.positioned import Positioned


class DrawObject(ABC, Positioned, Margined):
    def __init__(self, *args, **kwargs):
        self._show = True
        self._clipping_area = ClippingArea(pdf=None, draw_object=self)
        super().__init__(*args, **kwargs)

    @property
    def clipping_area(self):
        return self._clipping_area

    @property
    def show(self):
        return self._show

    @show.setter
    def show(self, val):
        if not isinstance(val, bool):
            raise TypeError(f"show.value must be of type bool not{type(val)}")
        self._show = val

    @abstractmethod
    def get_relative_x2(self):
        raise NotImplementedError()

    @abstractmethod
    def get_relative_y2(self):
        raise NotImplementedError()

    def get_height(self):
        return self.top_margin + self.get_relative_y2() - self.relative_y + self.bottom_margin

    def get_width(self):
        return self.left_margin + self.get_relative_x2() - self.relative_x + self.right_margin

    @abstractmethod
    def draw(self, pdf):
        raise NotImplementedError()

    def clipped_draw(self, pdf):
        self.clipping_area.pdf = pdf
        self.clipping_area.draw()

    def get_relative_position(self):
        return {'relative_x': self.relative_x, 'relative_y': self.relative_y}

    def get_margins(self):
        return {'left_margin': self.left_margin, 'top_margin': self.top_margin, 'right_margin': self.right_margin,
                'bottom_margin': self.bottom_margin}


class ClippingArea:
    def __init__(self, pdf, draw_object, left_margin=10, right_margin=0, top_margin=10):
        self.pdf = pdf
        self.draw_object = draw_object
        self.left_margin = left_margin
        self.right_margin = right_margin
        self.top_margin = top_margin

    @property
    def row_width(self):
        if not self.pdf:
            raise AttributeError('set pdf first!')
        return self.pdf.w - self.pdf.l_margin - self.pdf.r_margin - self.left_margin - self.right_margin

    def row_height(self):
        if not self.pdf:
            raise AttributeError('set pdf first!')
        return self.draw_object.get_height()

    def _prepare_page(self):
        self.pdf.translate_page_margins()
        self.pdf.translate(self.left_margin, self.top_margin)

    def _add_page(self):
        self.pdf.add_page()
        self._prepare_page()

    def _draw_with_clip(self, index):
        with self.pdf.saved_state():
            self.pdf.clip_rect(-1, -5, self.row_width + 1.14, self.row_height())
            self.pdf.translate(index * -self.row_width, 0)
            self.draw_object.draw(self.pdf)

    def get_number_of_rows(self):
        return int(ceil(self.draw_object.get_width() / self.row_width))

    def draw(self):
        self.pdf.translate(self.left_margin, self.top_margin)
        for index in range(self.get_number_of_rows()):
            if index != 0:
                self.pdf.translate(0, self.draw_object.get_height())
            if self.pdf.absolute_y > self.pdf.h - self.pdf.b_margin:
                self._add_page()
            self._draw_with_clip(index)



from pathlib import Path

from musurgia.fractaltree.fractaltree import FractalTree
from musurgia.pdf.pdf import Pdf
from musurgia.pdf.text import PageText
from musurgia.unittest import TestCase

path = Path(__file__)


def make_ft():
    ft = FractalTree(value=20)
    ft.add_layer()
    ft.get_children()[0].add_layer()
    ft.get_children()[-1].add_layer()
    return ft


class TestGenerateLayerSegmentedLine(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf(orientation='l')

    def test_draw(self):
        def change_ft_graphic():
            ft.graphic.unit = unit
            ft.graphic.add_labels(lambda node: round(float(node.value), 2), placement='above', font_size=6,
                                  bottom_margin=2, left_margin=0.5)
            ft.graphic.add_labels(lambda node: node.fractal_order, placement='below', font_size=6, top_margin=1,
                                  left_margin=0.5)

            ft.graphic.change_segment_attributes(bottom_margin=5)

        def change_layer_graphic(layer_graphic):
            for segment in layer_graphic.segments:
                segment.straight_line.add_label(round(float(segment.node.value), 2), font_size=6, bottom_margin=3)

        unit = 20
        with self.file_path(path, 'draw', 'pdf') as pdf_path:
            ft = make_ft()
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h', first_label=-1, unit=unit)
            self.pdf.draw_ruler('v')
            self.pdf.translate(15, 10)
            pt = PageText('Some Title', v_position='center', font_weight='bold', font_size=12)
            pt.draw(self.pdf)
            self.pdf.translate(0, 10)
            change_ft_graphic()
            ft.graphic.draw(self.pdf)
            self.pdf.translate(0, ft.graphic.get_height() + 10)
            segmented_line = ft.generate_layer_segmented_line(layer_number=2, unit=unit)
            change_layer_graphic(segmented_line)
            segmented_line.segments[0].start_mark_line.add_text_label('blabla', placement='left', right_margin=1)
            segmented_line.draw(self.pdf)
            self.pdf.write(pdf_path)



from unittest import TestCase

from musurgia.fractaltree.fractaltree import FractalTree


class Test(TestCase):
    def setUp(self) -> None:
        self.ft = FractalTree(proportions=[1, 2, 3], tree_permutation_order=[3, 1, 2], value=10)

    def test_1(self):
        self.ft.add_layer()
        self.ft.get_children()[1].add_layer()

        selected = self.ft.get_children()[1].get_children()[0]
        self.assertEqual(selected.__next__.name, '2.2')



import os

from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def test_1(self):
        fm = FractalMusic(tempo=60, proportions=[1, 2, 3, 4], tree_permutation_order=[3, 1, 4, 2], quarter_duration=100)
        fm.midi_generator.midi_range = [60, 79]
        fm.add_layer()
        partial_fm = fm.get_leaves()[3]
        partial_fm.add_layer()
        for leaf in partial_fm.traverse():
            leaf.chord.add_lyric(leaf.fractal_order)
            leaf.chord.add_words(leaf.midi_generator.midi_range)
            # leaf.chord.add_words(leaf.midi_generator.directions, relative_y=30)
        # print([leaf.fractal_order for leaf in partial_fm.traverse_leaves()])
        score = TreeScoreTimewise()
        v = partial_fm.get_simple_format(0).to_stream_voice(1)
        v.add_to_score(score, 1)

        v = partial_fm.get_simple_format().__deepcopy__().to_stream_voice(1)
        v.add_to_score(score, 2)

        partial_fm.reduce_children(condition=lambda child: child.fractal_order > 2)
        v = partial_fm.get_simple_format().__deepcopy__().to_stream_voice(1)
        v.add_to_score(score, 3)

        xml_path = path + '_test_1.xml'
        score.write(xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)

    # def test_3(self):
    #     fm = FractalMusic(proportions=[1, 2, 3, 4], tree_permutation_order=[3, 1, 4, 2], duration=100)
    #     fm.midi_generator.midi_range = [60, 79]
    #     fm.midi_generator.directions = [1, 1, -1]
    #     fm.add_layer()
    #     print(fm.children_generated_midis)
    #     print(fm.children_fractal_values)
    #     fm.reduce_children(lambda fractal_tree: fractal_tree.fractal_order == 2)
    #     print(fm.children_fractal_values)
    #     print(fm.midi_generator.proportions)
    #     print(fm.midi_generator.directions)
    #     print(fm.children_generated_midis)
    #     print(fm.proportions)
    #     fm.add_layer()
    #     print([leaf.fractal_order for leaf in fm.traverse_leaves()])

    # print([leaf.fractal_order for leaf in fm.traverse_leaves()])
    #
    # fm.reduce_leaves(lambda leaf: leaf.fractal_order == 2)
    # print([leaf.midi_value for leaf in fm.traverse_leaves()])

    # def test_4(self):
    #     permutation_order = [8, 11, 7, 12, 10, 13, 9, 4, 1, 3, 6, 2, 5]
    #     fm = FractalMusic(tree_permutation_order=permutation_order, duration=900, proportions=list(range(1, 14)))
    #     fm.midi_generator.midi_range = [48, 84]
    #     fm.midi_generator.directions = [1, 1, -1, -1]
    #     fm.add_layer()
    #     part_fm = fm.get_children()[6]
    #     part_fm.add_layer()
    #     print([fractal_tree.fractal_order for fractal_tree in part_fm.get_children()])
    #     part_fm.reduce_children(lambda fractal_tree: fractal_tree.fractal_order < 5)
    #     print([fractal_tree.fractal_order for fractal_tree in part_fm.get_children()])
    #     print([round(float(duration), 2) for duration in part_fm.simple_format.durations])

    # fm.reduce_leaves(lambda leaf: leaf.fractal_order == 2)

    def test_5(self):
        score = TreeScoreTimewise()

        fm = FractalMusic(tempo=60, proportions=[1, 2, 3, 4], tree_permutation_order=[3, 1, 4, 2], quarter_duration=20)
        fm.midi_generator.midi_range = [60, 79]
        fm.midi_generator.microtone = 4
        fm.add_layer()

        for child in fm.get_children():
            child.chord.add_lyric(child.fractal_order)

        simple_format = fm.get_simple_format(1)
        v = simple_format.to_stream_voice(1)
        v.add_to_score(score, 1)

        fm.reduce_children(lambda child: child.fractal_order in [1])

        simple_format = fm.get_simple_format(1)
        v = simple_format.to_stream_voice(1)
        v.add_to_score(score, 2)

        fm.add_layer()

        for leaf in fm.traverse_leaves():
            leaf.chord.add_lyric(leaf.fractal_order)
            # leaf.chord.add_words(leaf.midi_generator.midi_range[1])

        simple_format = fm.get_simple_format(2)
        v = simple_format.to_stream_voice(1)
        v.add_to_score(score, 3)

        text_path = path + '_test_5.txt'
        fm.write_infos(text_path)
        self.assertCompareFiles(actual_file_path=text_path)

        xml_path = path + '_test_5.xml'
        score.max_division = 7
        score.write(xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)

    def test_6(self):
        fm = FractalMusic(reading_direction='vertical', tempo=120, value=1.5, proportions=(1, 2, 3, 4, 5, 6, 7),
                          tree_permutation_order=(2, 6, 4, 1, 3, 7, 5), multi=(1, 4))
        fm.midi_generator.midi_range = (62, 70)
        fm.add_layer()
        # fm.generate_children(number_of_children=1)
        fm.reduce_children(lambda child: child.fractal_order < 7)
        result = fm.get_children()[0].midi_generator.midi_range
        # not [62, 70] because of directions ([-1, ...])
        expected = [70, 62]
        self.assertEqual(expected, result)



import os

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.fm = FractalMusic(proportions=[1, 2, 3], tree_permutation_order=[3, 1, 2], tempo=60, quarter_duration=20)

    def test_1(self):
        self.fm.midi_generator.midi_range = [60, 72]
        self.fm.add_layer()
        self.fm.get_children()[0].midi_value = 80
        self.fm.add_layer()
        for node in self.fm.traverse():
            if node.fractal_order is not None:
                node.chord.add_lyric(node.fractal_order)
        score = self.fm.get_score()
        score.max_division = 7

        text_path = path + '_test_1.txt'
        self.fm.write_infos(text_path)
        self.assertCompareFiles(actual_file_path=text_path)

        xml_path = path + '_test_1.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)

    def test_2(self):
        self.fm.midi_generator.midi_range = [60, 72]
        self.fm.midi_generator.set_directions(1, -1, 1)
        self.fm.add_layer()
        self.fm.get_children()[0].midi_value = 80

        split_nodes = self.fm.get_children()[0].split(1, 1)
        split_nodes[1].midi_value = 0

        self.fm.add_layer()

        for node in self.fm.traverse():
            if node.fractal_order is not None:
                node.chord.add_lyric(node.fractal_order)
        score = self.fm.get_score()
        score.max_division = 7

        text_path = path + '_test_2.txt'
        self.fm.write_infos(text_path)
        self.assertCompareFiles(actual_file_path=text_path)

        xml_path = path + '_test_2.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)



from fpdf import FPDF
from fpdf.php import sprintf

from musurgia.pdf.line import HorizontalRuler, VerticalRuler
from musurgia.pdf.pdfunit import PdfUnit
from musurgia.pdf.text import PageText


class SavedState:
    def __init__(self, pdf):
        self.pdf = pdf

    def __enter__(self):
        self.pdf._push_state()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pdf._pop_state()


class PrepareDrawObject:
    def __init__(self, pdf, draw_object):
        self.pdf = pdf
        self.draw_object = draw_object

    def __enter__(self):
        self.pdf._push_state()
        self.pdf.translate(self.draw_object.relative_x, self.draw_object.relative_y)
        self.pdf.translate(self.draw_object.left_margin, self.draw_object.top_margin)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pdf._pop_state()


class Pdf(FPDF):

    def __init__(self, r_margin=10, t_margin=10, l_margin=10, b_margin=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.r_margin = r_margin
        self.t_margin = t_margin
        self.l_margin = l_margin
        self.b_margin = b_margin
        self._absolute_positions = {}
        self.add_page()

        self._state = []

        self.set_font("Helvetica", "", 10)

    def _pop_state(self):
        self._state.pop()
        self._out(sprintf('Q\n'))

    def _push_state(self):
        self._state.append('push')
        self._out(sprintf('q\n'))

    @property
    def absolute_positions(self):
        return self._absolute_positions

    @property
    def absolute_x(self):
        return self._absolute_positions[self.page][0]

    @property
    def absolute_y(self):
        return self._absolute_positions[self.page][1]

    @property
    def k(self):
        return PdfUnit.get_k()

    @k.setter
    def k(self, val):
        pass

    def add_space(self, val):
        self.y += val

    def add_page(self):
        super().add_page(orientation=self.cur_orientation)
        self._absolute_positions[self.page] = [0, 0]

    def prepare_draw_object(self, draw_object):
        pdo = PrepareDrawObject(self, draw_object=draw_object)
        return pdo

    def reset_font(self):
        self._out(sprintf('BT /F%d %.2f Tf ET',
                          self.current_font['i'],
                          self.font_size_pt))

    def reset_position(self):
        self.translate(-self.absolute_x, -self.absolute_y)

    def clip_rect(self, x, y, w, h):
        self._out(sprintf('%.2f %.2f %.2f %.2f re W n',
                          x * self.k, (self.h - y) * self.k,
                          w * self.k, -h * self.k))

    def draw_page_numbers(self, **kwargs):
        for page in self.pages:
            self.page = page
            self.reset_position()
            page_number = PageText(page, **kwargs)
            page_number.draw(self)

    def saved_state(self):
        ss = SavedState(self)
        return ss

    def translate(self, dx, dy):
        if not self._state:
            self._absolute_positions[self.page][0] += dx
            self._absolute_positions[self.page][1] += dy
        dx, dy = dx * self.k, dy * self.k
        self._out(sprintf('1.0 0.0 0.0 1.0 %.2F %.2F cm',
                          dx, -dy))

    def translate_page_margins(self):
        self.translate(self.l_margin, self.t_margin)

    def draw_ruler(self, mode='h', unit=10, first_label=0, show_first_label=False, label_show_interval=1):
        if mode in ['h', 'horizontal']:
            length = self.w - self.l_margin - self.r_margin
            ruler = HorizontalRuler(length=length, unit=unit, first_label=first_label,
                                    show_first_label=show_first_label, label_show_interval=label_show_interval)
        elif mode in ['v', 'vertical']:
            length = self.h - self.t_margin - self.b_margin
            ruler = VerticalRuler(length=length, unit=unit, first_label=first_label,
                                  show_first_label=show_first_label, label_show_interval=label_show_interval)
        else:
            raise AttributeError()
        ruler.draw(self)

    def write(self, path):
        # if self.show_page_number:
        #     self.draw_page_number()
        self.output(path, 'F')



from musurgia.pdf.labeled import Labeled
from musurgia.pdf.newdrawobject import DrawObject


class DrawObjectContainer(DrawObject, Labeled):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._draw_objects = []

    def add_draw_object(self, draw_object):
        if not isinstance(draw_object, DrawObject):
            raise TypeError(draw_object)
        self._draw_objects.append(draw_object)
        return draw_object

    @property
    def draw_objects(self):
        return self._draw_objects


class DrawObjectRow(DrawObjectContainer):

    def get_relative_x2(self):
        return self.relative_x + sum([do.get_width() for do in self.draw_objects])

    def get_relative_y2(self):
        return self.relative_y + max([do.get_height() for do in self.draw_objects])

    def draw(self, pdf):
        with pdf.prepare_draw_object(self):
            self.draw_above_text_labels(pdf)
            self.draw_left_text_labels(pdf)
            for do in self.draw_objects:
                do.draw(pdf)
                pdf.translate(do.get_width(), 0)
            if self.below_text_labels:
                pdf.translate(0, 2)
                self.draw_below_text_labels(pdf)


class DrawObjectColumn(DrawObjectContainer):
    def get_relative_x2(self):
        return self.relative_x + max([do.get_width() for do in self.draw_objects])

    def get_relative_y2(self):
        return self.relative_y + sum([do.get_height() for do in self.draw_objects])

    def draw(self, pdf):
        with pdf.prepare_draw_object(self):
            self.draw_above_text_labels(pdf)
            self.draw_left_text_labels(pdf)
            for do in self.draw_objects:
                do.draw(pdf)
                pdf.translate(0, do.get_height())
            if self.below_text_labels:
                pdf.translate(0, 2)
                self.draw_below_text_labels(pdf)


from quicktions import Fraction

from musurgia.random import Random


class Interpolation(object):
    def __init__(self, start, end, duration=None, key=None, grid=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._start = None
        self._end = None
        self._duration = None
        self._grid = None

        self.start = start
        self.end = end
        self.duration = duration
        self.key = key
        self.grid = grid

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, val):
        self._start = val

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, val):
        self._end = val

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, val):
        if val and not isinstance(val, Fraction):
            val = Fraction(val)
        self._duration = val

    @property
    def grid(self):
        return self._grid

    @grid.setter
    def grid(self, val):
        self._grid = val

    def _check_grid(self, val):
        if self.grid:
            return round(val / self.grid) * self.grid
        else:
            return val

    def __call__(self, x):
        self._check_x(x)
        output = Fraction(Fraction(x * (self.end - self.start)), self.duration) + self.start
        output = self._check_grid(output)
        if self.key:
            return self.key(output)
        return output

    def _check_x(self, x):
        if x < 0 or x > self.duration:
            raise ValueError()


class InterpolationGroup(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._interpolations = []

    def add_interpolation(self, interpolation):
        if not isinstance(interpolation, Interpolation):
            raise TypeError
        self._interpolations.append(interpolation)

    def add_section(self, start, end, duration):
        interpolation = Interpolation(start, end, duration)
        self.add_interpolation(interpolation)

    def __call__(self, x):
        temp_x = x
        for interpolation in self._interpolations:
            try:
                return interpolation.__call__(temp_x)
            except ValueError:
                temp_x -= interpolation.duration

        raise ValueError(x)


class RandomInterpolation(Interpolation):
    def __init__(self, periodicity=None, seed=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._random = Random(periodicity=periodicity, seed=seed)

    @Interpolation.start.setter
    def start(self, val):
        if not hasattr(val, '__iter__'):
            val = [val]

        self._start = val
        self._check_pool_length()

    @Interpolation.end.setter
    def end(self, val):
        if not hasattr(val, '__iter__'):
            val = [val]
        self._end = val
        self._check_pool_length()

    def _check_pool_length(self):
        if self.start and self.end:
            if len(self.start) != len(self.end):
                return ValueError('start and end must be of the same length.')

    def _get_pool(self, x):
        pools = self.start + self.end
        pool_size = len(self.start)
        first_index = int(round(Interpolation(start=0, end=pool_size, duration=self.duration)(x)))
        pool = pools[first_index:first_index + pool_size]

        return pool

    def __call__(self, x):
        self._check_x(x)
        self._random.pool = self._get_pool(x)
        output = self._random.__next__()
        if self.key:
            return self.key(output)
        return output

    def __deepcopy__(self, memodict={}):
        copied = self.__class__(start=self.start, end=self.end, duration=self.duration, key=self.key)
        copied._random = self._random.__deepcopy__()
        return copied



from pathlib import Path

import matplotlib as mpl
from matplotlib.afm import AFM

from musurgia.unittest import TestCase

afm_path = Path(mpl.get_data_path(), 'fonts', 'afm', 'ptmr8a.afm')


def make_afm_path_dictionary():
    def check_entry():
        old_afm = output.get((family, weight, style))
        if old_afm is not None:
            old_header = old_afm._header
            new_header = afm._header
            diff = set(old_header) ^ set(new_header)
            if diff == {b'CapHeight'}:
                if new_header.get(b'CapHeight'):
                    return True
                #
                # print('old:', old_header.get(b'CapHeight'))
                # print('new:', new_header.get(b'CapHeight'))
            elif diff == set():
                return False
            else:
                raise AttributeError(
                    f'{family},  {weight}, {style} already in dict: {old_afm} differnce: {diff}')
        else:
            return True

    output = {}
    directory = Path(mpl.get_data_path(), 'fonts', 'afm')
    for file in directory.iterdir():
        afm_path = file
        with afm_path.open('rb') as fh:
            afm = AFM(fh)
        family = afm.get_familyname()
        weight = afm.get_weight().lower()
        if afm.get_angle() < 0:
            style = 'italic'
        else:
            style = 'regular'
        if check_entry():
            output[family, weight, style] = afm

    return output


class Test(TestCase):
    def test_afm_dict(self):
        actual = sorted(list(make_afm_path_dictionary().keys()))
        expected = [
            ('Computer Modern', 'medium', 'italic'),
            ('Computer Modern', 'medium', 'regular'),
            ('Courier', 'bold', 'italic'),
            ('Courier', 'bold', 'regular'),
            ('Courier', 'medium', 'italic'),
            ('Courier', 'medium', 'regular'),
            ('Helvetica', 'bold', 'italic'),
            ('Helvetica', 'bold', 'regular'),
            ('Helvetica', 'light', 'italic'),
            ('Helvetica', 'light', 'regular'),
            ('Helvetica', 'medium', 'italic'),
            ('Helvetica', 'medium', 'regular'),
            ('ITC Avant Garde Gothic', 'book', 'italic'),
            ('ITC Avant Garde Gothic', 'book', 'regular'),
            ('ITC Avant Garde Gothic', 'demi', 'italic'),
            ('ITC Avant Garde Gothic', 'demi', 'regular'),
            ('ITC Bookman', 'demi', 'italic'),
            ('ITC Bookman', 'demi', 'regular'),
            ('ITC Bookman', 'light', 'italic'),
            ('ITC Bookman', 'light', 'regular'),
            ('ITC Zapf Chancery', 'medium', 'italic'),
            ('ITC Zapf Dingbats', 'medium', 'regular'),
            ('New Century Schoolbook', 'bold', 'italic'),
            ('New Century Schoolbook', 'bold', 'regular'),
            ('New Century Schoolbook', 'medium', 'italic'),
            ('New Century Schoolbook', 'roman', 'regular'),
            ('Palatino', 'bold', 'italic'),
            ('Palatino', 'bold', 'regular'),
            ('Palatino', 'medium', 'italic'),
            ('Palatino', 'roman', 'regular'),
            ('Symbol', 'medium', 'regular'),
            ('Times', 'bold', 'italic'),
            ('Times', 'bold', 'regular'),
            ('Times', 'medium', 'italic'),
            ('Times', 'roman', 'regular'),
            ('Utopia', 'bold', 'italic'),
            ('Utopia', 'bold', 'regular'),
            ('Utopia', 'regular', 'italic'),
            ('Utopia', 'regular', 'regular')]
        self.assertEqual(expected, actual)

    def test_afm_dict_families(self):
        actual = list(dict.fromkeys([key[0] for key in make_afm_path_dictionary().keys()]))
        expected = ['New Century Schoolbook',
                    'Times',
                    'ITC Bookman',
                    'Helvetica',
                    'ITC Avant Garde Gothic',
                    'Palatino',
                    'Computer Modern',
                    'Symbol',
                    'ITC Zapf Dingbats',
                    'Utopia',
                    'Courier',
                    'ITC Zapf Chancery']
        self.assertEqual(expected, actual)

    def test_load_afm(self):
        afm = make_afm_path_dictionary()['Times', 'medium', 'italic']
        actual = afm.get_familyname()
        expected = 'Times'
        self.assertEqual(expected, actual)

    def test_width_height(self):
        afm = make_afm_path_dictionary()['Helvetica', 'bold', 'italic']
        actual = afm.string_width_height('What the heck?')
        expected = (7370.0, 741)
        self.assertEqual(expected, actual)
